import pytest
from faker import Faker

from wbfdm.factories.instruments import InstrumentFactory
from wbfdm.models import (
    Instrument,
    InstrumentClassificationThroughModel,
    RelatedInstrumentThroughModel,
)

fake = Faker()


@pytest.mark.django_db
class TestMergeInstrument:
    @pytest.fixture()
    def merged_instrument(self):
        return InstrumentFactory.create()

    @pytest.fixture()
    def main_instrument(self):
        return InstrumentFactory.create()

    def test_default(
        self,
        main_instrument,
        merged_instrument,
        exchange_factory,
        instrument_favorite_group_factory,
        related_instrument_through_model_factory,
        instrument_classification_through_model_factory,
    ):
        # Prepare data: add exchange
        main_exchange = main_instrument.exchange

        # Prepare data: Create Favorite group with the merged instrument within
        favorite_group = instrument_favorite_group_factory.create(instruments=[merged_instrument])

        # prepare data: related instruments
        main_related_instrument_rel = related_instrument_through_model_factory.create(instrument=main_instrument)
        merged_related_instrument_rel = related_instrument_through_model_factory.create(instrument=merged_instrument)
        reversed_related_instrument_rel = related_instrument_through_model_factory.create(
            related_instrument=merged_instrument
        )

        # prepare data: classification
        main_classification_rel = instrument_classification_through_model_factory.create(instrument=main_instrument)
        merged_classification_rel = instrument_classification_through_model_factory.create(
            instrument=merged_instrument
        )
        reversed_classification_rel = instrument_classification_through_model_factory.create(
            related_instruments=[merged_instrument]
        )

        main_instrument.merge(merged_instrument)

        # Test if the merged instrument is indeed deleted
        with pytest.raises(Instrument.DoesNotExist):
            merged_instrument.refresh_from_db()

        # Check Exchanges
        assert {main_instrument.exchange} == {
            main_exchange
        }  # check that the merged instrument exchange is forward to the main instrument

        # Check favorite Group
        assert list(favorite_group.instruments.all()) == [
            main_instrument
        ]  # Check that the favorite group with the merged instrument removed it and included the main instrument

        # Check related instruments
        assert (
            RelatedInstrumentThroughModel.objects.filter(
                instrument=main_instrument,
                related_instrument=merged_related_instrument_rel.related_instrument,
                is_primary=merged_related_instrument_rel.is_primary,
                related_type=merged_related_instrument_rel.related_type,
            ).exists()
        )  # Check that the related instrument with the merged instrument removed it and included the main instrument
        assert main_instrument.related_instruments.filter(
            id=main_related_instrument_rel.related_instrument.id
        ).exists()  # Check that the main instrument related instrument is still there
        assert RelatedInstrumentThroughModel.objects.filter(
            related_instrument=main_instrument, instrument=reversed_related_instrument_rel.instrument
        ).exists()  # Check that the reverse related instrument relationship has been forward to the main instrument

        # Check classification
        assert InstrumentClassificationThroughModel.objects.filter(
            instrument=main_instrument,
            classification=merged_classification_rel.classification,
            is_favorite=merged_classification_rel.is_favorite,
            reason=merged_classification_rel.reason,
            pure_player=merged_classification_rel.pure_player,
            top_player=merged_classification_rel.top_player,
            percent_of_revenue=merged_classification_rel.percent_of_revenue,
        ).exists()
        assert main_instrument.classifications.filter(
            id=main_classification_rel.classification.id
        ).exists()  # Check that the main instrument classification relationship is still there
        assert (
            InstrumentClassificationThroughModel.objects.filter(
                related_instruments=main_instrument, classification=reversed_classification_rel.classification
            ).exists()
        )  # Check that the reverse classificaiton instrument relationship has been forward to the main instrument



# Generated by Django 4.1.9 on 2023-07-11 08:24

import django.db.models.deletion
import wbfdm.models.fk_fields
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("wbfdm", "0002_rename_statements_instrumentlookup_financials_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="instrument",
            name="estimate_backend",
            field=wbfdm.models.fk_fields.CompositeForeignKey(
                null=True,
                null_if_equal=[],
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="wbfdm.instrumentlookup",
                to_fields={
                    "estimates": wbfdm.models.fk_fields.RawFieldValue(True),
                    "instrument_id": wbfdm.models.fk_fields.LocalFieldValue("id"),
                },
            ),
        ),
        migrations.AlterField(
            model_name="instrumentlookup",
            name="parameters",
            field=models.JSONField(blank=True, default=dict, null=True),
        ),
    ]



from datetime import date

import pandas as pd
from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from pandas.tseries.offsets import QuarterEnd, YearEnd
from tqdm import tqdm
from wbcore.contrib.io.models import ImportedObjectProviderRelationship, Provider

from wbfdm.contrib.dsws.client import Client
from wbfdm.models.instruments.instruments import Instrument

REPORT_TYPE_MAP = {"QTR": "Q", "TRI": "4M", "SAN": "6M", "ANN": "Y"}


class Controller:
    def __init__(self, client_username: str, client_password: str):
        self.provider = Provider.objects.get(key="refinitiv")
        self.client = Client(username=client_username, password=client_password)

    @classmethod
    def _wrap_identifier_in_bracklet(cls, perm_ids: list[str]) -> list[str]:
        return [f"<{perm_id}>" for perm_id in perm_ids]

    def get_start_and_end_date_for_interim(self, perm_id: str, end: date, start: date | None = None) -> date | None:
        if not start:
            start = end - QuarterEnd(1)
        if rel := ImportedObjectProviderRelationship.objects.filter(
            provider=self.provider,
            provider_identifier=perm_id,
            content_type=ContentType.objects.get_for_model(Instrument),
        ).first():
            instrument = rel.content_object
            fundamentals = instrument.fundamentals.order_by("-period__period_year", "-period__period_index")
            # Not an historical import, it's a bulk one, we need to get the next fiscal period to fetch
            if fundamentals.exists() and (fiscal_period := fundamentals.first().period):
                start = fiscal_period.calendar_period.end_time.date()

        return (start + QuarterEnd(0)).date(), (
            end + QuarterEnd(0) + QuarterEnd(1)
        ).date()  # We always refetch since the beginning of the quarter

    @classmethod
    def get_start_date_for_annual(cls, end: date, start: date | None = None) -> date:
        if not start:
            start = end
        return (start - YearEnd(1)).date(), (
            end + YearEnd(0)
        ).date()  # We always refetch since the beginning of the quarter

    def get_frequency(self, instrument: str) -> str:
        try:
            res = self.client.get_static_df([instrument], ["IBEFPD"])
            if not res.empty:
                return REPORT_TYPE_MAP[res.set_index("Instrument").loc[instrument, "IBEFPD"]]
        except Exception:
            return "Q"

    def fetch_perm_id(
        self,
        instrument_ric: str = None,
        instrument_isin: str = None,
        instrument_mnemonic: str = None,
        perm_id_symbols: list[str] = ["QPID", "IPID"],
    ) -> str | None:
        def _process_ticker(ticker):
            if not (df := self.client.get_static_df(tickers=[ticker], fields=perm_id_symbols)).empty:
                for perm_id in perm_id_symbols:
                    if identifier := df[perm_id].iloc[0]:
                        return identifier

        res = None
        if instrument_ric:
            res = _process_ticker(f"<{instrument_ric}>")
        if instrument_mnemonic and not res:
            res = _process_ticker(instrument_mnemonic)
        if instrument_isin and not res:
            res = _process_ticker(instrument_isin)
        return res

    def get_data(
        self,
        identifiers: list[str],
        fields: list[str],
        start: date | None = None,
        end: date | None = None,
        wrap_tickers_into_brackets: bool = True,
        **extra_client_kwargs,
    ) -> pd.DataFrame:
        df_list = []
        chunked_identifiers = self.client.get_chunked_list(identifiers, len(fields))

        if settings.DEBUG:
            chunked_identifiers = tqdm(chunked_identifiers, total=len(chunked_identifiers))

        frequency = extra_client_kwargs.pop("freq", "D")
        for perm_ids in chunked_identifiers:
            if wrap_tickers_into_brackets:
                perm_ids = self._wrap_identifier_in_bracklet(perm_ids)
            if start or end:
                dff = self.client.get_timeserie_df(
                    perm_ids, fields, start=start, end=end, freq=frequency, **extra_client_kwargs
                )
                df_list.append(dff)
            else:
                df_list.append(self.client.get_static_df(perm_ids, fields, **extra_client_kwargs))
        if df_list:
            return pd.concat(df_list, axis=0)
        return pd.DataFrame()

    def get_interim_fundamental_data(
        self,
        identifiers: list[str],
        annual_fields: list[str],
        initial_start: date | None = None,
        initial_end: date | None = None,
        wrap_tickers_into_brackets: bool = True,
        **extra_client_kwargs,
    ) -> pd.DataFrame:
        df_list = []
        interim_fields_map = {field + "A" if field[0:2] == "WC" else field: field for field in annual_fields}
        if settings.DEBUG:
            identifiers = tqdm(identifiers, total=len(identifiers))

        for instrument_perm_id in identifiers:
            start = initial_start  # copy start argument
            end = initial_end  # copy end argument

            start, end = self.get_start_and_end_date_for_interim(instrument_perm_id, end, start=start)
            if start and end:
                perm_ids = [instrument_perm_id]
                if wrap_tickers_into_brackets:
                    perm_ids = self._wrap_identifier_in_bracklet(perm_ids)
                # frequency = self.get_frequency(identifiers[0])
                df = self.client.get_timeserie_df(
                    perm_ids,
                    list(interim_fields_map.keys()),
                    start=start,
                    end=end,
                    **extra_client_kwargs,
                )
                if not df.empty:
                    df_list.append(df)
        if df_list:
            df = pd.concat(df_list, axis=0).dropna(how="all")
            df["period__period_interim"] = True
            return df.rename(columns=interim_fields_map)
        return pd.DataFrame()

    def get_annual_fundamental_data(
        self,
        identifiers: list[str],
        annual_fields: list[str],
        initial_start: date | None = None,
        initial_end: date | None = None,
        wrap_tickers_into_brackets: bool = True,
        **extra_client_kwargs,
    ) -> pd.DataFrame:
        df_list = []
        start, end = self.get_start_date_for_annual(initial_end, start=initial_start)
        chunked_identifiers = self.client.get_chunked_list(identifiers, len(annual_fields))
        if settings.DEBUG:
            chunked_identifiers = tqdm(chunked_identifiers, total=len(chunked_identifiers))
        for perm_ids in chunked_identifiers:
            if wrap_tickers_into_brackets:
                perm_ids = self._wrap_identifier_in_bracklet(perm_ids)
            df = self.client.get_timeserie_df(
                perm_ids,
                annual_fields,
                start=start,
                end=end,
                freq="Y",
                **extra_client_kwargs,
            )
            if not df.empty:
                df_list.append(df)

        if df_list:
            df = pd.concat(df_list, axis=0).dropna(how="all")
            df["WC05200"] = 4
            df["period__period_interim"] = False
            return df
        return pd.DataFrame()



import factory
from slugify import slugify

from wbfdm.models import InstrumentList, InstrumentListThroughModel


class InstrumentListFactory(factory.django.DjangoModelFactory):
    name = factory.Faker("word")
    identifier = factory.LazyAttribute(lambda o: slugify(o.name))
    instrument_list_type = None

    class Meta:
        django_get_or_create = ("identifier",)
        model = InstrumentList


class InstrumentListThroughModelFactory(factory.django.DjangoModelFactory):
    instrument_list = factory.SubFactory(InstrumentListFactory)
    instrument = factory.SubFactory("wbfdm.factories.instruments.InstrumentFactory")

    class Meta:
        django_get_or_create = ("instrument_list",)
        model = InstrumentListThroughModel



from datetime import date
from decimal import Decimal
from typing import Iterator

from wbcore.contrib.dataloader.dataloaders import Dataloader

from wbfdm.dataloaders.protocols import MarketDataProtocol
from wbfdm.dataloaders.types import MarketDataDict
from wbfdm.enums import MarketData
from wbfdm.models.instruments.instrument_prices import InstrumentPrice

MarketDataMap = {
    "SHARES_OUTSTANDING": "internal_outstanding_shares",
    "OPEN": "net_value",
    "CLOSE": "net_value",
    "HIGH": "net_value",
    "LOW": "net_value",
    "BID": "net_value",
    "ASK": "net_value",
    "VOLUME": "internal_volume",
    "MARKET_CAPITALIZATION": "market_capitalization",
}

DEFAULT_VALUES = [MarketData[name] for name in MarketDataMap.keys()]


def _cast_decimal_to_float(value: float | Decimal) -> float:
    if isinstance(value, Decimal):
        value = float(value)
    return value


class MarketDataDataloader(MarketDataProtocol, Dataloader):
    def market_data(
        self,
        values: list[MarketData] | None = [MarketData.CLOSE],
        from_date: date | None = None,
        to_date: date | None = None,
        exact_date: date | None = None,
        calculated: bool | None = None,
        **kwargs,
    ) -> Iterator[MarketDataDict]:
        """Get prices for instruments.

        Args:
            values (list[MarketData]): List of values to include in the results.
            from_date (date | None): The starting date for filtering prices. Defaults to None.
            to_date (date | None): The ending date for filtering prices. Defaults to None.
            frequency (Frequency): The frequency of the requested data

        Returns:
            Iterator[MarketDataDict]: An iterator of dictionaries conforming to the DailyValuationDict.
        """
        prices = InstrumentPrice.objects.filter(instrument__in=self.entities).annotate_market_data()  # type: ignore
        if calculated is not None:
            prices = prices.filter(calculated=calculated)
        else:
            prices = prices.filter_only_valid_prices()

        prices = prices.order_by("date")
        if not values:
            values = DEFAULT_VALUES
        values_map = {value.name: MarketDataMap[value.name] for value in values if value.name in MarketDataMap}
        if exact_date:
            prices = prices.filter(date=exact_date)
        else:
            if from_date:
                prices = prices.filter(date__gte=from_date)
            if to_date:
                prices = prices.filter(date__lte=to_date)
        for row in prices.filter_only_valid_prices().values(
            "date",
            "instrument",
            "calculated",
            *set(values_map.values()),
        ):
            external_id = row.pop("instrument")
            val_date = row.pop("date")
            if row:
                yield MarketDataDict(
                    id=f"{external_id}_{val_date}",
                    valuation_date=val_date,
                    instrument_id=external_id,
                    external_id=external_id,
                    source="wbfdm",
                    calculated=row["calculated"],
                    **{MarketData[k].value: _cast_decimal_to_float(row[v]) for k, v in values_map.items()},
                )



import time
from datetime import datetime

import pytest
from django.core.cache import cache as cache_layer
from django.utils import timezone
from faker import Faker
from jwt import encode
from requests.exceptions import ConnectionError

from ..client import MSCIClient, is_expired

fake = Faker()


class TestMSCIClient:
    @pytest.fixture()
    def jwt(self, expiry_date: datetime):
        payload = {"exp": int(expiry_date.timestamp()), "iat": int(timezone.now().timestamp())}
        return encode(payload, "secret", algorithm="HS256")

    @pytest.fixture()
    def msci_client(self):
        return MSCIClient("username", "password")

    @pytest.mark.parametrize("expiry_date", [fake.past_datetime()])
    def test_is_expired_true(self, expiry_date, jwt):
        assert is_expired(jwt)

    @pytest.mark.parametrize("expiry_date", [fake.future_datetime()])
    def test_is_expired_false(self, expiry_date, jwt):
        assert not is_expired(jwt)

    def test_oauth_token_not_in_cache(self, requests_mock, msci_client):
        access_token = fake.word()
        expiry = fake.pyint(min_value=60)
        requests_mock.post(MSCIClient.AUTH_SERVER_ULR, json={"access_token": access_token, "expires_in": expiry})
        t0 = time.time()
        res = msci_client.oauth_token
        assert res == access_token
        assert cache_layer.get("msci_oauth_token") == access_token
        assert int(cache_layer._expire_info.get(":1:msci_oauth_token")) == int(t0 + expiry)

    @pytest.mark.parametrize("expiry_date", [fake.future_datetime()])
    def test_oauth_token_in_cache_and_valid(self, requests_mock, msci_client, expiry_date, jwt):
        requests_mock.post(MSCIClient.AUTH_SERVER_ULR, json={"access_token": fake.word()})
        cache_layer.set("msci_oauth_token", jwt)
        assert msci_client.oauth_token == jwt

    @pytest.mark.parametrize("expiry_date", [fake.past_datetime()])
    def test_oauth_token_in_cache_but_expired(self, requests_mock, msci_client, expiry_date, jwt):
        requests_mock.post(MSCIClient.AUTH_SERVER_ULR, json={"access_token": fake.word()})
        cache_layer.set("msci_oauth_token", jwt)

        access_token = fake.word()
        requests_mock.post(
            MSCIClient.AUTH_SERVER_ULR, json={"access_token": access_token, "expires_in": fake.pyint(min_value=60)}
        )
        assert msci_client.oauth_token == access_token

    def test_oauth_token_connection_error(self, requests_mock, msci_client):
        cache_layer.clear()
        # keyerror
        requests_mock.post(MSCIClient.AUTH_SERVER_ULR, json={"access_token_typo": fake.word()})
        with pytest.raises(ConnectionError):
            t = msci_client.oauth_token  # noqa

        requests_mock.post(MSCIClient.AUTH_SERVER_ULR, status_code=500)
        with pytest.raises(ConnectionError):
            t = msci_client.oauth_token  # noqa



# Generated by Django 5.0.9 on 2024-11-06 13:20

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
        ("metric", "0001_initial"),
        ("wbfdm", "0027_remove_instrument_unique_ric_and_more"),
    ]

    operations = [
        migrations.RemoveConstraint(
            model_name="instrumentmetric",
            name="unique_instrument_metric",
        ),
        migrations.AddConstraint(
            model_name="instrumentmetric",
            constraint=models.UniqueConstraint(
                fields=("basket_content_type", "basket_id", "instrument", "date", "key"),
                name="unique_instrument_metric",
                nulls_distinct=False,
            ),
        ),
    ]



from django_celery_beat.models import IntervalSchedule, PeriodicTask
from django.db import DEFAULT_DB_ALIAS
from django.apps import apps as global_apps


def initialize_task(app_config, verbosity=2, interactive=True, using=DEFAULT_DB_ALIAS, apps=global_apps, **kwargs):
    PeriodicTask.objects.get_or_create(
        task="wbfdm.tasks.update_of_investable_universe_data",
        defaults={
            "name": "FDM: Periodic data update of investable universe",
            "interval": IntervalSchedule.objects.get_or_create(every=3, period=IntervalSchedule.HOURS)[0],
            "crontab": None,
        },
    )
    PeriodicTask.objects.get_or_create(
        task="wbfdm.tasks.synchronize_instruments_as_task",
        defaults={
            "name": "FDM: Synchronize Instrument",
            "interval": IntervalSchedule.objects.get_or_create(every=30, period=IntervalSchedule.MINUTES)[0],
            "crontab": None,
        },
    )
    PeriodicTask.objects.get_or_create(
        task="wbfdm.tasks.synchronize_exchanges_as_task",
        defaults={
            "name": "FDM: Synchronize Exchange",
            "interval": IntervalSchedule.objects.get_or_create(every=30, period=IntervalSchedule.MINUTES)[0],
            "crontab": None,
        },
    )



import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
from django.db.models import Count, Value
from django.db.models.expressions import OuterRef, Subquery
from django.db.models.functions import Coalesce
from django.shortcuts import get_object_or_404
from django.utils.functional import cached_property
from reversion.views import RevisionMixin
from wbcore import viewsets
from wbcore.cache.decorators import cache_table
from wbcore.filters import DjangoFilterBackend
from wbcore.permissions.permissions import InternalUserPermissionMixin
from wbcore.utils.date import get_next_day_timedelta

from wbfdm.filters import (
    ClassificationFilter,
    ClassificationTreeChartFilter,
    InstrumentClassificationThroughModelViewFilterSet,
)
from wbfdm.models import (
    Classification,
    ClassificationGroup,
    InstrumentClassificationThroughModel,
)
from wbfdm.serializers import (
    ClassificationGroupModelSerializer,
    ClassificationGroupRepresentationSerializer,
    ClassificationModelSerializer,
    ClassificationRepresentationSerializer,
    InstrumentClassificationThroughModelSerializer,
)
from wbfdm.viewsets.configs import (
    ChildClassificationParentClassificationTitleConfig,
    ClassificationButtonConfig,
    ClassificationClassificationGroupEndpointConfig,
    ClassificationClassificationGroupTitleConfig,
    ClassificationDisplayConfig,
    ClassificationEndpointConfig,
    ClassificationGroupButtonConfig,
    ClassificationGroupDisplayConfig,
    ClassificationGroupTitleConfig,
    ClassificationIcicleChartEndpointConfig,
    ClassificationIcicleChartTitleConfig,
    ClassificationInstrumentThroughInstrumentModelEndpointConfig,
    ClassificationParentClassificationEndpointConfig,
    ClassificationTitleConfig,
    ClassificationTreeChartEndpointConfig,
    ClassificationTreeChartTitleConfig,
    InstrumentClassificationThroughDisplayConfig,
    InstrumentClassificationThroughEndpointConfig,
    InstrumentClassificationThroughInstrumentModelEndpointConfig,
    InstrumentClassificationThroughTitleConfig,
)

from ..mixins import InstrumentMixin


class ClassificationRepresentationViewSet(InternalUserPermissionMixin, viewsets.RepresentationViewSet):
    IDENTIFIER = "wbfdm:classification"
    queryset = Classification.objects.all()
    serializer_class = ClassificationRepresentationSerializer

    filterset_class = ClassificationFilter
    ordering_fields = ["code_aggregated"]
    ordering = ["code_aggregated"]
    search_fields = ["code_aggregated", "name"]


class ClassificationGroupRepresentationViewSet(InternalUserPermissionMixin, viewsets.RepresentationViewSet):
    IDENTIFIER = "wbfdm:classificationgrouprepresentation"
    queryset = ClassificationGroup.objects.all()
    serializer_class = ClassificationGroupRepresentationSerializer

    filterset_fields = {"is_primary": ["exact"], "max_depth": ["gte", "exact", "lte"]}
    ordering_fields = ["name"]
    ordering = ["name"]
    search_fields = ["name"]


class ClassificationModelViewSet(InternalUserPermissionMixin, RevisionMixin, viewsets.ModelViewSet):
    IDENTIFIER = "wbfdm:classification"

    queryset = Classification.objects.select_related("parent", "group")
    serializer_class = ClassificationModelSerializer
    filterset_class = ClassificationFilter

    ordering_fields = (
        "name",
        "code_aggregated",
        "level_representation",
    )
    ordering = ("level_representation", "name")
    search_fields = ("name", "group__name", "parent__name", "code_aggregated")

    endpoint_config_class = ClassificationEndpointConfig
    display_config_class = ClassificationDisplayConfig
    title_config_class = ClassificationTitleConfig
    button_config_class = ClassificationButtonConfig

    @cached_property
    def next_valid_code_aggregated(self) -> str:
        group = None
        parent = None
        if group_id := self.kwargs.get("group_id", None):
            group = ClassificationGroup.objects.get(id=group_id)
        elif parent_id := self.kwargs.get("parent_id", None):
            parent = Classification.objects.get(id=parent_id)
            group = parent.group
        if group:
            return Classification.get_next_valid_code(group, parent=parent)
        return ""


class ClassificationClassificationGroupModelViewSet(ClassificationModelViewSet):
    endpoint_config_class = ClassificationClassificationGroupEndpointConfig
    title_config_class = ClassificationClassificationGroupTitleConfig

    def get_queryset(self):
        return super().get_queryset().filter(group=self.kwargs["group_id"], level=0)


class ChildClassificationParentClassificationModelViewSet(ClassificationModelViewSet):
    endpoint_config_class = ClassificationParentClassificationEndpointConfig
    title_config_class = ChildClassificationParentClassificationTitleConfig

    def get_queryset(self):
        parent = Classification.objects.get(id=self.kwargs["parent_id"])
        return super().get_queryset().filter(parent=self.kwargs["parent_id"], level=parent.level + 1)


class ClassificationGroupModelViewSet(InternalUserPermissionMixin, RevisionMixin, viewsets.ModelViewSet):
    IDENTIFIER = "wbfdm:classificationgroup"

    queryset = ClassificationGroup.objects.all()
    serializer_class = ClassificationGroupModelSerializer
    filterset_fields = {"is_primary": ["exact"], "max_depth": ["gte", "exact", "lte"]}
    ordering_fields = ("name",)
    ordering = ("name",)
    search_fields = ("name",)

    display_config_class = ClassificationGroupDisplayConfig
    title_config_class = ClassificationGroupTitleConfig
    button_config_class = ClassificationGroupButtonConfig


class AbstractClassificationChartView(viewsets.ChartViewSet):
    filter_backends = (DjangoFilterBackend,)
    queryset = Classification.objects.all()
    filterset_class = ClassificationTreeChartFilter

    @cached_property
    def aggregation_type(self) -> str:
        return self.request.GET.get("aggregation_type", "classification_count")

    @cached_property
    def classification_group(self) -> ClassificationGroup:
        return ClassificationGroup.objects.get(id=self.kwargs["group_id"])

    @cached_property
    def top_level(self) -> int:
        top_level = 0
        if top_classification_id := self.request.GET.get("top_classification", None):
            top_level = Classification.objects.get(id=top_classification_id).level
        return top_level

    def get_df(self, queryset) -> tuple[pd.DataFrame, list[str]]:
        _range = self.classification_group.max_depth + 1 - self.top_level
        df = pd.DataFrame(columns=[self.aggregation_type])
        level_representation = []
        if queryset.exists():
            df = pd.DataFrame(
                queryset.values(*[f"{'parent__' * height}name" for height in range(_range)], self.aggregation_type)
            ).fillna(0)

            level_representation = self.classification_group.get_levels_representation()[:_range]
            df.columns = [*level_representation, self.aggregation_type]
        return df, level_representation

    def get_queryset(self):
        return (
            super()
            .get_queryset()
            .filter(height=0, group=self.classification_group)
            .annotate(
                instrument_count=Coalesce(
                    Subquery(
                        InstrumentClassificationThroughModel.objects.filter(classification=OuterRef("pk"))
                        .values("classification")
                        .annotate(c=Count("classification"))
                        .values("c")[:1]
                    ),
                    0,
                ),
                classification_count=Value(1),
            )
        )

    def get_plotly(self, queryset):
        return go.Figure()


@cache_table(
    timeout=get_next_day_timedelta(),
    key_prefix=lambda view: f"{view.classification_group.id}_{view.aggregation_type}_{view.top_level}",
    periodic_caching=True,
    periodic_caching_view_kwargs=lambda: [
        {"group_id": group_id} for group_id in ClassificationGroup.objects.values_list("id", flat=True)
    ],
    periodic_caching_get_parameters=[
        {"aggregation_type": "classification_count"},
        {"aggregation_type": "instrument_count"},
    ],
)
class ClassificationTreeChartView(AbstractClassificationChartView):
    IDENTIFIER = "wbfdm:classificationgroup-tree"
    title_config_class = ClassificationTreeChartTitleConfig
    endpoint_config_class = ClassificationTreeChartEndpointConfig

    def get_plotly(self, queryset):
        df, level_representation = self.get_df(queryset)
        fig = px.treemap(df, path=[*level_representation[::-1]], values=self.aggregation_type, branchvalues="total")
        fig.update_traces(root_color="lightgrey", hovertemplate=f"{self.aggregation_type}: %{{value}}")
        fig.update_layout(margin=dict(t=25, l=25, r=25, b=25))
        return fig


@cache_table(
    timeout=get_next_day_timedelta(),
    key_prefix=lambda view: f"{view.classification_group.id}_{view.aggregation_type}_{view.top_level}",
    periodic_caching=True,
    periodic_caching_view_kwargs=lambda: [
        {"group_id": group_id} for group_id in ClassificationGroup.objects.values_list("id", flat=True)
    ],
    periodic_caching_get_parameters=[
        {"aggregation_type": "classification_count"},
        {"aggregation_type": "instrument_count"},
    ],
)
class ClassificationIcicleChartView(AbstractClassificationChartView):
    IDENTIFIER = "wbfdm:classificationgroup-iciclechart"
    title_config_class = ClassificationIcicleChartTitleConfig
    endpoint_config_class = ClassificationIcicleChartEndpointConfig

    def get_plotly(self, queryset):
        df, level_representation = self.get_df(queryset)
        fig = px.icicle(df, path=[*level_representation[::-1]], values=self.aggregation_type)
        fig.update_traces(
            root_color="lightgrey", tiling=dict(orientation="v"), hovertemplate=f"{self.aggregation_type}: %{{value}}"
        )
        fig.update_layout(margin=dict(t=25, l=25, r=25, b=25))
        return fig


class InstrumentClassificationThroughModelViewSet(InternalUserPermissionMixin, viewsets.ModelViewSet):
    serializer_class = InstrumentClassificationThroughModelSerializer
    queryset = InstrumentClassificationThroughModel.objects.select_related(
        "instrument",
        "classification",
    ).prefetch_related("related_instruments", "tags")
    search_fields = ("classification__name", "classification__code_aggregated")
    ordering_fields = ordering = ["classification__name"]

    filterset_class = InstrumentClassificationThroughModelViewFilterSet
    display_config_class = InstrumentClassificationThroughDisplayConfig
    endpoint_config_class = InstrumentClassificationThroughEndpointConfig
    title_config_class = InstrumentClassificationThroughTitleConfig


class InstrumentClassificationThroughInstrumentModelViewSet(
    InstrumentMixin, InstrumentClassificationThroughModelViewSet
):
    endpoint_config_class = InstrumentClassificationThroughInstrumentModelEndpointConfig
    search_fields = ["classification__computed_str"]
    ordering_fields = ordering = ["classification__computed_str"]

    def get_queryset(self):
        return super().get_queryset().filter(instrument__in=self.instrument.get_family())


class ClassificationInstrumentThroughInstrumentModelViewSet(InstrumentClassificationThroughModelViewSet):
    search_fields = ["instrument__computed_str"]
    ordering_fields = ordering = ["instrument__computed_str"]

    endpoint_config_class = ClassificationInstrumentThroughInstrumentModelEndpointConfig

    def get_queryset(self):
        classification = get_object_or_404(Classification, id=self.kwargs["classification_id"])
        return (
            super()
            .get_queryset()
            .filter(classification__in=classification.get_descendants(include_self=True).values("id"))
        )



from django.apps import AppConfig
from django.conf import settings
from django.core.checks import Warning, register


@register()
def check_qa_db_available(app_configs, **kwargs):
    warnings = []

    if "qa" not in settings.DATABASES:
        warnings.append(
            Warning(
                "A database with the name qa needs to be installed in order to run QA",
            )
        )

    return warnings


class QAAppConfig(AppConfig):
    name = "wbfdm.contrib.qa"
    verbose_name = "Quantitative Analytics (QA)"



from contextlib import suppress
from decimal import Decimal

import pandas as pd
from celery import shared_task
from django.db import models, transaction
from django.db.models import (
    Case,
    DecimalField,
    Exists,
    ExpressionWrapper,
    F,
    OuterRef,
    Q,
    QuerySet,
    Subquery,
    Value,
    When,
)
from django.db.models.signals import post_save
from django.dispatch import receiver
from wbcore.contrib.currency.models import CurrencyFXRates
from wbcore.contrib.io.mixins import ImportMixin
from wbcore.models import DynamicDecimalField, DynamicFloatField, DynamicModel, WBModel
from wbcore.signals import pre_merge

from wbfdm.analysis.financial_analysis.financial_statistics_analysis import (
    FinancialStatistics,
)
from wbfdm.import_export.handlers.instrument_price import InstrumentPriceImportHandler

from .mixin.financials_computed import InstrumentPriceComputedMixin


class ValidPricesQueryset(QuerySet):
    def filter_only_valid_prices(self) -> QuerySet:
        """
        Filter the queryset to remove duplicate in case calculated and non-calculated prices are present for the same date/product/type
        """
        return self.annotate(
            real_price_exists=Exists(
                self.filter(
                    instrument=OuterRef("instrument"),
                    date=OuterRef("date"),
                    calculated=False,
                )
            )
        ).filter(Q(calculated=False) | (Q(real_price_exists=False) & Q(calculated=True)))

    def annotate_market_data(self):
        base_qs = ValidPricesQueryset(self.model, using=self._db)

        return self.annotate(
            currency_fx_rate_to_usd_rate=F("currency_fx_rate_to_usd__value"),
            calculated_outstanding_shares=Subquery(
                base_qs.filter(instrument=OuterRef("instrument"), calculated=True, date=OuterRef("date")).values(
                    "outstanding_shares"
                )[:1]
            ),
            internal_outstanding_shares=ExpressionWrapper(
                Case(  # Annotate the parent security if exists
                    When(
                        outstanding_shares__isnull=False,
                        then=F("outstanding_shares"),
                    ),
                    default=F("calculated_outstanding_shares"),
                ),
                output_field=DecimalField(max_digits=4),
            ),
            internal_market_capitalization=Case(
                When(
                    market_capitalization__isnull=True,
                    then=ExpressionWrapper(
                        F("internal_outstanding_shares") * F("net_value"),
                        output_field=DecimalField(max_digits=4),
                    ),
                ),
                default=ExpressionWrapper(F("market_capitalization"), output_field=DecimalField(max_digits=4)),
            ),
            internal_market_capitalization_usd=F("internal_market_capitalization") / F("currency_fx_rate_to_usd_rate"),
            calculated_volume=Subquery(
                base_qs.filter(instrument=OuterRef("instrument"), calculated=True, date=OuterRef("date")).values(
                    "volume"
                )[:1]
            ),
            internal_volume=ExpressionWrapper(
                Case(  # Annotate the parent security if exists
                    When(
                        volume__isnull=False,
                        then=F("volume"),
                    ),
                    default=F("calculated_volume"),
                ),
                output_field=models.FloatField(),
            ),
        )

    def annotate_base_data(self):
        return self.annotate(
            currency_fx_rate_to_usd_rate=F("currency_fx_rate_to_usd__value"),
            market_capitalization_usd=ExpressionWrapper(
                F("market_capitalization") / F("currency_fx_rate_to_usd_rate"), output_field=DecimalField(max_digits=4)
            ),
            net_value_usd=F("net_value") / F("currency_fx_rate_to_usd_rate"),
            gross_value_usd=F("gross_value") / F("currency_fx_rate_to_usd_rate"),
            volume_usd=ExpressionWrapper(
                F("volume") * F("net_value") / F("currency_fx_rate_to_usd_rate"), output_field=DecimalField()
            ),
            volume_50d_usd=ExpressionWrapper(
                F("volume_50d") * F("net_value") / F("currency_fx_rate_to_usd_rate"), output_field=DecimalField()
            ),
            volume_200d_usd=ExpressionWrapper(
                F("volume_200d") * F("net_value") / F("currency_fx_rate_to_usd_rate"), output_field=DecimalField()
            ),
        )

    def annotate_security_data(self):
        return self.annotate(
            security=Case(  # Annotate the parent security if exists
                When(instrument__parent__isnull=False, then=F("instrument__parent")),
                default=F("instrument"),
            ),
            security_instrument_type_key=Case(  # Annotate the parent security if exists
                When(
                    instrument__parent__isnull=False,
                    then=F("instrument__parent__instrument_type__key"),
                ),
                default=F("instrument__instrument_type__key"),
            ),
        )

    def annotate_all(self):
        return self.annotate_market_data().annotate_security_data().annotate_base_data()


class InstrumentPriceManager(models.Manager):
    def __init__(self, with_annotation: bool = False, *args, **kwargs):
        self.with_annotation = with_annotation
        super().__init__(*args, **kwargs)

    def get_queryset(self) -> ValidPricesQueryset:
        qs = ValidPricesQueryset(self.model)
        if self.with_annotation:
            qs = qs.annotate_all()
        return qs

    def filtered_by_instruments(self, instrument_queryset, *other_instruments):
        return self.filter(models.Q(instrument__in=instrument_queryset) | models.Q(instrument__in=other_instruments))

    def filter_only_valid_prices(self) -> QuerySet:
        return self.get_queryset().filter_only_valid_prices()

    def annotate_market_data(self) -> QuerySet:
        return self.get_queryset().annotate_market_data()

    def annotate_base_data(self) -> QuerySet:
        return self.get_queryset().annotate_base_data()

    def annotate_security_data(self) -> QuerySet:
        return self.get_queryset().annotate_security_data()

    def annotate_all(self) -> QuerySet:
        return self.get_queryset().annotate_all()


class AnnotatedInstrumentPriceManager(InstrumentPriceManager):
    def get_queryset(self):
        return super().get_queryset().annotate_market_data().annotate_base_data().annotate_security_data()


class InstrumentPrice(
    ImportMixin,
    InstrumentPriceComputedMixin,
    DynamicModel,
    WBModel,
):
    import_export_handler_class = InstrumentPriceImportHandler
    # Base fields
    instrument = models.ForeignKey(
        to="wbfdm.Instrument",
        related_name="prices",
        on_delete=models.PROTECT,
        limit_choices_to=models.Q(children__isnull=True),
        verbose_name="Instrument",
        blank=True,
        null=True,
    )
    date = models.DateField(verbose_name="Date")
    calculated = models.BooleanField(default=False, verbose_name="Is Calculated")

    net_value = models.DecimalField(max_digits=16, decimal_places=6, verbose_name="Value (Net)")
    gross_value = DynamicDecimalField(
        max_digits=16,
        decimal_places=6,
        verbose_name="Value (Gross)",
    )  # TODO: I think we need to remove this field that is not really used here.

    outstanding_shares = DynamicDecimalField(
        decimal_places=4,
        max_digits=16,
        verbose_name="Outstanding Shares",
        help_text="The amount of outstanding share for this instrument",
    )
    outstanding_shares_consolidated = DynamicDecimalField(
        decimal_places=4,
        max_digits=16,
        verbose_name="Outstanding Shares (Consolidated)",
        help_text="The amount of outstanding share for this instrument",
    )
    ########################################################
    #                   ASSET STATISTICS                   #
    ########################################################

    volume = models.FloatField(
        default=0.0,
        verbose_name="Volume",
        help_text="The Volume of the Asset on the price date of the Asset.",
    )

    volume_50d = DynamicFloatField(
        verbose_name="Average Volume (50 Days)",
        help_text="The Average Volume of the Asset over the last 50 days from the price date of the Asset.",
    )

    volume_200d = models.FloatField(
        null=True,
        blank=True,
        verbose_name="Average Volume (200 Days)",
        help_text="The Average Volume of the Asset over the last 200 days from the price date of the Asset.",
    )
    market_capitalization = models.FloatField(
        null=True,
        blank=True,
        verbose_name="Market Capitalization",
        help_text="The Market Capitalization of the Asset the price date of the Asset.",
    )
    market_capitalization_consolidated = models.FloatField(
        null=True,
        blank=True,
        verbose_name="Market Capitalization (Consolidated)",
        help_text="the consolidated market value of a company in local currency.",
    )

    currency_fx_rate_to_usd = models.ForeignKey(
        to="currency.CurrencyFXRates",
        related_name="instrument_prices",
        on_delete=models.PROTECT,
        blank=True,
        null=True,
        verbose_name="Instrument Currency Rate",
        help_text="Rate to between instrument currency and USD",
    )

    # Statistics
    lock_statistics = models.BooleanField(
        default=False,
        help_text="If True, a save will not override the beta, correlation and sharpe ratio",
    )
    sharpe_ratio = models.FloatField(blank=True, null=True, verbose_name="Sharpe Ratio")
    correlation = models.FloatField(blank=True, null=True, verbose_name="Correlation")
    beta = models.FloatField(blank=True, null=True, verbose_name="Beta")
    annualized_daily_volatility = models.FloatField(blank=True, null=True, verbose_name="Annualized Volatility")

    created = models.DateTimeField(auto_now_add=True, verbose_name="Created")
    modified = models.DateTimeField(
        verbose_name="Modified",
        auto_now=True,
    )
    # custom_beta_180d = DynamicFloatField(verbose_name="Custom Beta (180 days)")
    # custom_beta_1y = DynamicFloatField(verbose_name="Custom Beta (1 Years)")
    # custom_beta_2y = DynamicFloatField(verbose_name="Custom Beta (2 Years)")
    # custom_beta_3y = DynamicFloatField(verbose_name="Custom Beta (3 Years)")
    # custom_beta_5y = DynamicFloatField(verbose_name="Custom Beta (4 Years)")
    #
    # # Performances fields
    # performance_1d = DynamicDecimalField(
    #     verbose_name="Performance 1D", help_text="Performance 1D", max_digits=16, decimal_places=6
    # )
    # performance_7d = DynamicDecimalField(
    #     verbose_name="Performance (1W)", help_text="Performance 7 days rolling", max_digits=16, decimal_places=6
    # )
    # performance_30d = DynamicDecimalField(
    #     verbose_name="Performance (1M)", help_text="Performance 30 days rolling", max_digits=16, decimal_places=6
    # )
    # performance_90d = DynamicDecimalField(
    #     verbose_name="Performance (1Q)", help_text="Performance 90 days rolling", max_digits=16, decimal_places=6
    # )
    # performance_365d = DynamicDecimalField(
    #     verbose_name="Performance (1Y)", help_text="Performance 365 days rolling", max_digits=16, decimal_places=6
    # )
    # performance_ytd = DynamicDecimalField(
    #     verbose_name="Performance (YTD)", help_text="Performance Year-to-date", max_digits=16, decimal_places=6
    # )
    # performance_inception = DynamicDecimalField(
    #     verbose_name="Performance (Inception)",
    #     help_text="Performance since inception",
    #     max_digits=16,
    #     decimal_places=6,
    # )

    objects = InstrumentPriceManager()
    annotated_objects = InstrumentPriceManager(with_annotation=True)

    class Meta:
        verbose_name = "Instrument Price"
        verbose_name_plural = "Instrument Prices"
        constraints = [
            models.CheckConstraint(
                check=~models.Q(date__week_day__in=[1, 7]),
                name="%(app_label)s_%(class)s_weekday_constraint",
            ),
            models.UniqueConstraint(fields=["instrument", "date", "calculated"], name="unique_price"),
        ]
        indexes = [
            models.Index(
                name="fdm_instrumentprice_base_idx",
                fields=["calculated", "date", "instrument"],
            ),
            models.Index(
                name="fdm_instrumentprice_idx1",
                fields=["calculated", "instrument"],
            ),
            models.Index(
                name="fdm_instrumentprice_idx2",
                fields=["instrument"],
            ),
        ]

    @property
    def _net_value_usd(self):
        if self.currency_fx_rate_to_usd:
            return getattr(self, "net_value_usd", self.net_value * self.currency_fx_rate_to_usd.value)

    def save(self, *args, **kwargs):
        if not self.currency_fx_rate_to_usd:
            with suppress(CurrencyFXRates.DoesNotExist):
                self.currency_fx_rate_to_usd = CurrencyFXRates.objects.get(
                    date=self.date, currency=self.instrument.currency
                )

        if self.market_capitalization_consolidated is None:
            self.market_capitalization_consolidated = self.market_capitalization
        super().save(*args, **kwargs)

    def __str__(self):
        return f"{self.instrument.name}: {self.net_value} {self.date:%d.%m.%Y}"

    @property
    def previous_price(self):
        """Returns the previous instrument prices if one exists or None

        Returns:
            instrument.InstrumentPrice | None -- Previous InstrumentPrice
        """
        try:
            return InstrumentPrice.objects.filter(
                instrument=self.instrument, date__lt=self.date, calculated=self.calculated
            ).latest("date")
        except InstrumentPrice.DoesNotExist:
            return None

    @property
    def next_price(self):
        """Returns the next instrument prices if one exists or None

        Returns:
            instrument.InstrumentPrice | None -- Next InstrumentPrice
        """
        try:
            return self.instrument.prices.filter(date__gt=self.date, calculated=self.calculated).earliest("date")
        except InstrumentPrice.DoesNotExist:
            return None

    @property
    def shares(self):
        """Returns the number of shares of a instrument

        The number of shares are the sum of all customer trades

        Returns:
            int -- Shares
        """
        return self.instrument.total_shares(self.date)

    @property
    def valid_outstanding_shares(self):
        prices = self.instrument.prices.filter(date=self.date, outstanding_shares__isnull=False).order_by("calculated")
        return prices.last().outstanding_shares if prices.exists() else self.shares

    @property
    def nominal_value(self):
        """Returns the nominal value of a instrument

        The nominal value is the number of current shares multiplied by the share price of a instrument

        Returns:
            int -- Nominal Value
        """
        return self.instrument.nominal_value(self.date)

    def fill_market_capitalization(self):
        if self.market_capitalization is None:
            with suppress(InstrumentPrice.DoesNotExist):
                self.market_capitalization = (
                    self.instrument.valuations.filter(date__lt=self.date, market_capitalization__isnull=False)
                    .latest("date")
                    .market_capitalization
                )

    def compute_and_update_statistics(self, min_period: int = 20):
        df = (
            pd.DataFrame(
                InstrumentPrice.objects.filter_only_valid_prices()
                .filter(instrument=self.instrument, date__lte=self.date)
                .values_list("date", "net_value", "volume"),
                columns=["date", "net_value", "volume"],
            )
            .set_index("date")
            .sort_index()
        )
        prices_df = df["net_value"]
        if not prices_df.empty and prices_df.shape[0] >= min_period:
            financial_statistics = FinancialStatistics(prices_df)
            min_date = prices_df.index.min()
            if risk_free_rate := self.instrument.primary_risk_instrument:
                risk_free_rate_df = (
                    pd.DataFrame(
                        risk_free_rate.valuations.filter(date__gte=min_date, date__lte=self.date).values_list(
                            "date", "net_value"
                        ),
                        columns=["date", "net_value"],
                    )
                    .set_index("date")
                    .sort_index()["net_value"]
                )
                if sharpe_ratio := financial_statistics.get_sharpe_ratio(risk_free_rate_df):
                    self.sharpe_ratio = sharpe_ratio
            if benchmark := self.instrument.primary_benchmark:
                benchmark_df = (
                    pd.DataFrame(
                        benchmark.valuations.filter(date__gte=min_date, date__lte=self.date).values_list(
                            "date", "net_value"
                        ),
                        columns=["date", "net_value"],
                    )
                    .set_index("date")
                    .sort_index()["net_value"]
                )
                if (beta := financial_statistics.get_beta(benchmark_df)) is not None:
                    self.beta = beta
                if (correlation := financial_statistics.get_correlation(benchmark_df)) is not None:
                    self.correlation = correlation

            self.annualized_daily_volatility = financial_statistics.get_annualized_daily_volatility()
        if not (volume_df := df["volume"]).empty:
            self.volume_50d = volume_df.tail(50).mean()
            self.volume_200d = volume_df.tail(200).mean()

    @classmethod
    def subquery_closest_value(
        cls,
        field_name,
        val_date=None,
        date_name="date",
        instrument_pk_name="instrument__pk",
        date_lookup="lte",
        order_by="-date",
        calculated_filter_value=False,
    ):
        index_filter_params = {}
        if calculated_filter_value is not None:
            index_filter_params["calculated"] = calculated_filter_value
        if not val_date and date_name:
            index_filter_params[f"date__{date_lookup}"] = models.OuterRef(date_name)
        elif val_date:
            index_filter_params[f"date__{date_lookup}"] = val_date
        index_filter_params["instrument"] = models.OuterRef(instrument_pk_name)
        qs = cls.objects.filter(**index_filter_params).filter(**{f"{field_name}__isnull": False})
        return models.Subquery(qs.order_by(order_by).values(field_name)[:1])

    @classmethod
    def annotate_sum_shares(cls, queryset, val_date, date_key="date"):
        """
        Efficient way to annotate sum of shares without destroying indexing
        """
        return queryset.annotate(
            sum_shares_calculated=InstrumentPrice.subquery_closest_value(
                "outstanding_shares",
                date_name=date_key,
                instrument_pk_name="pk",
                date_lookup="lte",
                order_by="-date",
                calculated_filter_value=True,
            ),  # We get the last price whose outstanding_shares is not none and date below date_key
            sum_shares_real=InstrumentPrice.subquery_closest_value(
                "outstanding_shares",
                val_date=val_date,
                date_name=date_key,
                instrument_pk_name="pk",
                date_lookup="exact",
                calculated_filter_value=False,
            ),
            sum_shares=Case(
                When(sum_shares_real__isnull=False, then=F("sum_shares_real")),
                When(sum_shares_calculated__isnull=False, then=F("sum_shares_calculated")),
                default=Value(Decimal(0)),
                output_field=DecimalField(),
            ),
        )

    @classmethod
    def get_endpoint_basename(cls):
        return "wbfdm:price"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{instrument} {{date}}"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbfdm:price-list"


@receiver(post_save, sender="currency.CurrencyFXRates")
def rate_creation(sender, instance, created, raw, **kwargs):
    if not raw and created:
        transaction.on_commit(lambda: update_currency_fx_rate_from_created_rate.delay(instance.id))


@receiver(pre_merge, sender="wbfdm.Instrument")
def pre_merge_instrument(sender: models.Model, merged_object, main_object, **kwargs):
    """
    Simply reassign the prices of the merged instrument to the main instrument if they don't already exist for that day, otherwise simply delete them
    """
    merged_object.prices.annotate(
        already_exists=Exists(
            InstrumentPrice.objects.filter(
                calculated=OuterRef("calculated"), instrument=main_object, date=OuterRef("date")
            )
        )
    ).filter(already_exists=True).delete()
    merged_object.prices.update(instrument=main_object)


@shared_task(queue="portfolio")
def update_currency_fx_rate_from_created_rate(rate_id):
    currency_rate = CurrencyFXRates.objects.get(id=rate_id)
    for price in InstrumentPrice.objects.filter(
        instrument__currency=currency_rate.currency, date=currency_rate.date, currency_fx_rate_to_usd__isnull=True
    ):
        price.currency_fx_rate_to_usd = currency_rate
        price.save()



import pytest
from django.contrib.contenttypes.models import ContentType

from wbfdm.models import Instrument

from ..dto import Metric
from ..models import InstrumentMetric
from ..registry import backend_registry


@pytest.mark.django_db
class TestInstrumentMetric:
    def test_save(self, instrument_metric):
        assert instrument_metric.basket_repr == str(instrument_metric.basket)

    def test_update_or_create_from_metric(self, weekday, instrument_factory):
        i1 = instrument_factory.create()
        i2 = instrument_factory.create()
        i3 = instrument_factory.create()
        key = "key"
        basket_content_type_id = ContentType.objects.get_for_model(i1).id
        level1_dto_metric = Metric(
            basket_id=i1.id,
            basket_content_type_id=basket_content_type_id,
            key=key,
            metrics={"a": "a"},
            date=weekday,
        )
        level1_dto_metric = Metric(
            basket_id=i2.id,
            basket_content_type_id=basket_content_type_id,
            key=key,
            metrics={"b": "b"},
            date=weekday,
            dependency_metrics=[level1_dto_metric],
        )
        base_dto_metric = Metric(
            basket_id=i3.id,
            basket_content_type_id=basket_content_type_id,
            key=key,
            metrics={"c": "c"},
            date=weekday,
            dependency_metrics=[level1_dto_metric],
        )
        InstrumentMetric.update_or_create_from_metric(base_dto_metric)

        level2_metric = InstrumentMetric.objects.get(
            basket_id=i1.id, basket_content_type_id=basket_content_type_id, date=weekday, key=key
        )
        level1_metric = InstrumentMetric.objects.get(
            basket_id=i2.id, basket_content_type_id=basket_content_type_id, date=weekday, key=key
        )
        base_metric = InstrumentMetric.objects.get(
            basket_id=i3.id, basket_content_type_id=basket_content_type_id, date=weekday, key=key
        )

        assert level2_metric.metrics == {"a": "a"}
        assert level2_metric.parent_metric == level1_metric

        assert level1_metric.metrics == {"b": "b"}
        assert level1_metric.parent_metric == base_metric

        assert base_metric.metrics == {"c": "c"}

    def test_annotate_with_metrics(self, instrument_metric):
        qs = Instrument.objects.filter(id=instrument_metric.basket_id)
        qs = InstrumentMetric.annotate_with_metrics(
            qs, backend_registry._metric_key_map[instrument_metric.key], Instrument, val_date=instrument_metric.date
        )
        for field in instrument_metric.metrics.keys():
            qs_field = instrument_metric.key + "_" + field
            res = list(qs.values_list(qs_field, flat=True))[0]
            assert res == instrument_metric.metrics[field]



import pytest

from wbfdm.models.instruments import (
    Classification,
    ClassificationGroup,
    InstrumentClassificationThroughModel,
)


@pytest.mark.django_db
class TestClassificationGroupModel:
    def test_init(self, classification_group):
        assert classification_group.id is not None

    def test_primary(self, classification_group_factory):
        g1 = classification_group_factory.create()
        assert g1.is_primary
        g2 = ClassificationGroup.objects.create(is_primary=True, name="Other group")
        g1.refresh_from_db()
        assert g2.is_primary
        assert not g1.is_primary

    def test_get_levels_representation(self, classification_group_factory, classification_factory):
        group = classification_group_factory.create(max_depth=2)
        c0 = classification_factory.create(name="c0", group=group, level=0, parent=None)
        c1, created = Classification.objects.update_or_create(group=group, level=1, parent=c0, defaults={"name": "c1"})
        c2, created = Classification.objects.update_or_create(group=group, level=2, parent=c1, defaults={"name": "c2"})
        assert set(group.get_levels_representation()) == {
            c0.level_representation,
            c1.level_representation,
            c2.level_representation,
        }


@pytest.mark.django_db
class TestClassificationModel:
    def test_init(self, classification_factory):
        classification = classification_factory.create()
        assert classification.id is not None

    def test_str(self, classification_factory):
        classification = classification_factory.create()
        assert str(classification) == classification.computed_str

    def test_get_instruments(self, classification_factory, instrument_factory, classification_group):
        ind1 = classification_factory.create(group=classification_group)
        subind1 = classification_factory.create(parent=ind1, group=classification_group)
        subind2 = classification_factory.create(parent=ind1, group=classification_group)

        e1 = instrument_factory.create(classifications=(subind1,))
        e2 = instrument_factory.create(classifications=(subind2,))
        InstrumentClassificationThroughModel.objects.filter(instrument=e2, classification=subind2).update(
            is_favorite=True
        )

        assert ind1.get_classified_instruments().count() == 2
        assert subind1.get_classified_instruments().first() == e1
        assert subind2.get_classified_instruments().first() == e2

        assert ind1.get_classified_instruments(only_favorites=True).count() == 1
        assert subind1.get_classified_instruments(only_favorites=True).first() is None
        assert subind2.get_classified_instruments(only_favorites=True).first() == e2

    @pytest.mark.parametrize("classification_group__code_level_digits", [1, 2, 3, 4])
    def test_get_next_valid_code(self, classification_factory, classification_group):
        parent_classification = classification_factory.create(group=classification_group, code_aggregated=None)
        assert parent_classification.code_aggregated == f"{1:0{classification_group.code_level_digits}}"
        c1 = Classification.objects.filter(parent=parent_classification, group=classification_group).first()
        assert (
            c1.code_aggregated
            == f"{1:0{classification_group.code_level_digits}}" + f"{1:0{classification_group.code_level_digits}}"
        )
        c2 = classification_factory.create(
            parent=parent_classification, group=classification_group, code_aggregated=None
        )
        assert (
            c2.code_aggregated
            == f"{1:0{classification_group.code_level_digits}}" + f"{2:0{classification_group.code_level_digits}}"
        )

    def test_get_default_level_representation(self, classification_factory, classification_group_factory):
        classification_group = classification_group_factory.create(max_depth=2)

        # Parent Classification has a default "Level 0" level representation.
        parent_classification = classification_factory.create(group=classification_group)
        assert parent_classification.level_representation == "Level 0"

        # If no siblings, default is the roman style representation of parent level + 1.
        c1 = Classification.objects.filter(parent=parent_classification, group=classification_group).first()
        assert c1.level_representation == "Level I"

        # C1 is the parent of C2, we write our own level representation for level 2 classifications.
        c2 = Classification.objects.filter(parent=c1, group=classification_group).first()
        assert c2.level_representation == "Level II"
        c2.level_representation = "Level B"
        c2.save()

        # If we now create a new level 2 classification, it will take the same representation of its siblings.
        c22 = classification_factory.create(parent=c1, group=classification_group)
        assert c22.level_representation == "Level B"



# Generated by Django 5.0.6 on 2024-07-04 07:38

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("currency", "0001_initial"),
        ("geography", "0001_initial"),
        ("io", "0007_alter_exportsource_query_params"),
        ("tags", "0001_initial"),
        ("wbfdm", "0022_instrument_cusip_option_open_interest_20d_and_more"),
    ]

    operations = [
        migrations.AddConstraint(
            model_name="instrument",
            constraint=models.UniqueConstraint(
                condition=models.Q(("is_security", True)), fields=("refinitiv_identifier_code",), name="unique_ric"
            ),
        ),
        migrations.AddConstraint(
            model_name="instrument",
            constraint=models.UniqueConstraint(
                condition=models.Q(("is_security", True)), fields=("refinitiv_mnemonic_code",), name="unique_rmc"
            ),
        ),
        migrations.AddConstraint(
            model_name="instrument",
            constraint=models.UniqueConstraint(
                condition=models.Q(("is_security", True)), fields=("isin",), name="unique_isin"
            ),
        ),
        migrations.AddConstraint(
            model_name="instrument",
            constraint=models.UniqueConstraint(
                condition=models.Q(("is_security", True)), fields=("sedol",), name="unique_sedol"
            ),
        ),
        migrations.AddConstraint(
            model_name="instrument",
            constraint=models.UniqueConstraint(
                condition=models.Q(("is_security", True)), fields=("valoren",), name="unique_valoren"
            ),
        ),
        migrations.AddConstraint(
            model_name="instrument",
            constraint=models.UniqueConstraint(
                condition=models.Q(("is_security", True)), fields=("cusip",), name="unique_cusip"
            ),
        ),
    ]



from contextlib import suppress
from itertools import batched
from typing import Iterator

import pypika as pk
from django.db import ProgrammingError, connections
from pypika import functions as fn
from pypika.analytics import RowNumber
from pypika.enums import SqlTypes
from pypika.terms import LiteralValue
from wbcore.contrib.dataloader.dataloaders import Dataloader
from wbcore.contrib.dataloader.utils import dictfetchall

from wbfdm.dataloaders.protocols import OfficersProtocol
from wbfdm.dataloaders.types import OfficerDataDict


class RKDOfficersDataloader(OfficersProtocol, Dataloader):
    def officers(
        self,
    ) -> Iterator[OfficerDataDict]:
        lookup = {k: v for k, v in self.entities.values_list("dl_parameters__officers__parameters", "id")}

        # Define tables
        designation = pk.Table("RKDFndCmpOffTitleChg")
        officer = pk.Table("RKDFndCmpOfficer")
        temp_codes = pk.Table("#rkd_codes")

        # Build the query
        query = (
            pk.MSSQLQuery.select(
                fn.Concat(designation.Code, "-", RowNumber().orderby(officer.OfficerRank)).as_("id"),
                designation.Code.as_("external_identifier"),
                designation.Title.as_("position"),
                fn.Concat(
                    officer.Prefix,
                    " ",
                    officer.FirstName,
                    " ",
                    officer.LastName,
                    pk.Case().when(officer.Suffix.isnull(), "").else_(fn.Concat(", ", officer.Suffix)),
                ).as_("name"),
                officer.Age.as_("age"),
                officer.Sex.as_("sex"),
                fn.Cast(designation.DesgStartDt, SqlTypes.DATE).as_("start"),
            )
            .from_(designation)
            .join(officer)
            .on((designation.Code == officer.Code) & (designation.OfficerID == officer.Officerid))
            .where(designation.Code.isin([LiteralValue("select code from #rkd_codes")]))
            .where(designation.DesgEndDt.isnull())
            .orderby(officer.OfficerRank)
        )

        with connections["qa"].cursor() as cursor:
            # Create and populate temporary table
            with suppress(ProgrammingError):
                cursor.execute(
                    pk.MSSQLQuery.create_table(temp_codes).columns(pk.Column("code", SqlTypes.INTEGER)).get_sql()
                )
                for batch in batched(lookup.keys(), 1000):
                    placeholders = ",".join(map(lambda x: f"('{x}')", batch))
                    cursor.execute(f"insert into #rkd_codes values {placeholders};")

            cursor.execute(query.get_sql())

            for row in dictfetchall(cursor, OfficerDataDict):
                row["instrument_id"] = lookup[row["external_identifier"]]
                yield row

            # Clean up temporary table
            cursor.execute(pk.MSSQLQuery.drop_table(temp_codes).get_sql())



from django.db.models import Prefetch, Q
from django.db.models.expressions import F
from django.db.models.query import QuerySet
from django.utils.functional import cached_property
from rest_framework import filters
from wbcore import filters as wb_filters
from wbcore import serializers as wb_serializers
from wbcore import viewsets
from wbcore.contrib.tags.serializers import TagSerializerMixin
from wbcore.permissions.permissions import InternalUserPermissionMixin

from wbfdm.filters.instruments import BaseClassifiedInstrumentFilterSet
from wbfdm.models import (
    Classification,
    ClassificationGroup,
    Instrument,
    InstrumentFavoriteGroup,
    RelatedInstrumentThroughModel,
)
from wbfdm.models.instruments import (
    InstrumentClassificationRelatedInstrument,
    InstrumentClassificationThroughModel,
)
from wbfdm.preferences import get_default_classification_group
from wbfdm.serializers import (
    ClassifiableInstrumentRepresentationSerializer,
    ClassificationRepresentationSerializer,
    InstrumentClassificationRelatedInstrumentModelSerializer,
    InstrumentClassificationRelatedInstrumentRepresentationSerializer,
    InstrumentFavoriteGroupModelSerializer,
    InstrumentFavoriteGroupRepresentationSerializer,
    ReadOnlyInstrumentFavoriteGroupModelSerializer,
    RelatedInstrumentThroughInstrumentModelSerializer,
)
from wbfdm.viewsets.configs.display import ClassifiedInstrumentDisplayConfig
from wbfdm.viewsets.configs.titles import ClassifiedInstrumentTitleConfig

from ..configs import (
    ClassificationInstrumentRelatedInstrumentDisplayConfig,
    ClassificationInstrumentRelatedInstrumentEndpointConfig,
    ClassifiedInstrumentEndpointConfig,
    InstrumentFavoriteGroupDisplayConfig,
    InstrumentFavoriteGroupEndpointConfig,
    InstrumentFavoriteGroupTitleConfig,
    RelatedInstrumentThroughInstrumentDisplayConfig,
    RelatedInstrumentThroughInstrumentEndpointConfig,
)
from ..mixins import InstrumentMixin


class InstrumentClassificationRelatedInstrumentRepresentationViewSet(viewsets.RepresentationViewSet):
    queryset = InstrumentClassificationRelatedInstrument.objects.all()
    serializer_class = InstrumentClassificationRelatedInstrumentRepresentationSerializer

    def get_queryset(self):
        queryset = super().get_queryset()
        if pk := self.kwargs.get("classified_instrument_id", None):
            queryset = queryset.filter(classified_instrument_id=pk)
        return queryset


class InstrumentClassificationRelatedInstrumentModelViewSet(viewsets.ModelViewSet):
    queryset = InstrumentClassificationRelatedInstrument.objects.all()
    serializer_class = InstrumentClassificationRelatedInstrumentModelSerializer
    display_config_class = ClassificationInstrumentRelatedInstrumentDisplayConfig
    endpoint_config_class = ClassificationInstrumentRelatedInstrumentEndpointConfig

    def get_queryset(self):
        queryset = super().get_queryset()
        if pk := self.kwargs.get("classified_instrument_id", None):
            queryset = queryset.filter(classified_instrument_id=pk)
        return queryset


class InstrumentFavoriteGroupRepresentationViewSet(InternalUserPermissionMixin, viewsets.RepresentationViewSet):
    IDENTIFIER = "wbfdm:favoritegroup"
    filter_backends = (filters.OrderingFilter, filters.SearchFilter)
    ordering_fields = ordering = ("name",)
    search_fields = ("name", "instruments__name", "owner__computed_str")
    queryset = InstrumentFavoriteGroup.objects.all()
    serializer_class = InstrumentFavoriteGroupRepresentationSerializer


class InstrumentFavoriteGroupModelViewSet(InternalUserPermissionMixin, viewsets.ModelViewSet):
    queryset = InstrumentFavoriteGroup.objects.all()
    serializer_class = InstrumentFavoriteGroupModelSerializer

    ordering_fields = ("name", "owner__computed_str", "public")
    ordering = ("name",)
    search_fields = ("name", "instruments__name", "owner__computed_str")

    filterset_fields = {"instruments": ["exact"], "owner": ["exact"], "public": ["exact"]}

    display_config_class = InstrumentFavoriteGroupDisplayConfig
    title_config_class = InstrumentFavoriteGroupTitleConfig
    endpoint_config_class = InstrumentFavoriteGroupEndpointConfig

    @cached_property
    def is_owner(self) -> bool:
        try:
            group = InstrumentFavoriteGroup.objects.get(id=self.kwargs["pk"])
            return group.owner == self.request.user.profile
        except (KeyError, InstrumentFavoriteGroup.DoesNotExist):
            return False

    def get_serializer_class(self):
        if self.is_owner:
            return InstrumentFavoriteGroupModelSerializer
        return ReadOnlyInstrumentFavoriteGroupModelSerializer

    def get_queryset(self):
        qs = InstrumentFavoriteGroup.objects.all()
        if not self.request.user.is_superuser:
            qs = qs.filter(Q(owner=self.request.user.profile) | Q(public=True))
        return qs.select_related("owner").prefetch_related(
            Prefetch("instruments", queryset=Instrument.objects.filter(favorite_groups__isnull=False).distinct())
        )


class RelatedInstrumentThroughInstrumentModelViewSet(
    InstrumentMixin, InternalUserPermissionMixin, viewsets.ModelViewSet
):
    serializer_class = RelatedInstrumentThroughInstrumentModelSerializer
    queryset = RelatedInstrumentThroughModel.objects.select_related(
        "related_instrument",
        "instrument",
    )

    search_fields = ("related_instrument__computed_str",)
    ordering_fields = ["is_primary"]
    ordering = ["-is_primary"]

    filterset_fields = {"is_primary": ["exact"], "related_instrument": ["exact"], "related_type": ["exact"]}
    display_config_class = RelatedInstrumentThroughInstrumentDisplayConfig
    endpoint_config_class = RelatedInstrumentThroughInstrumentEndpointConfig

    def get_queryset(self):
        return super().get_queryset().filter(instrument=self.instrument)


def get_classified_instrument_serializer_class(classification_grouo_id: int) -> type:
    """
    Unwrap defined serializer class and inject the metric fields into a new class
    """

    group = ClassificationGroup.objects.get(id=classification_grouo_id)

    attrs = dict()
    serializer_fields = [
        "id",
        "instrument",
        "_instrument",
        "classification",
        "_classification",
        "is_favorite",
        "tags",
        "_tags",
    ]
    for field_name in group.get_fields_names(sep="_"):
        base_field_name = f"classification_{field_name}"
        # representation_field_name = f"_classification_{field_name}"
        attrs[base_field_name] = wb_serializers.CharField(read_only=True)
        serializer_fields.append(base_field_name)

    class BaseClassifiedInstrumentModelSerializer(TagSerializerMixin, wb_serializers.ModelSerializer):
        _instrument = ClassifiableInstrumentRepresentationSerializer(source="instrument")
        _classification = ClassificationRepresentationSerializer(source="classification", label_key="{{name}}")

        class Meta:
            model = InstrumentClassificationThroughModel
            fields = read_only_fields = serializer_fields

    serializer_class = type("ClassifiedInstrumentModelSerializer", (BaseClassifiedInstrumentModelSerializer,), attrs)

    return serializer_class


class ClassifiedInstrumentModelViewSet(InternalUserPermissionMixin, viewsets.ModelViewSet):
    queryset = InstrumentClassificationThroughModel.objects.all()
    display_config_class = ClassifiedInstrumentDisplayConfig
    title_config_class = ClassifiedInstrumentTitleConfig
    endpoint_config_class = ClassifiedInstrumentEndpointConfig

    search_fields = ("instrument__computed_str",)

    def get_ordering_fields(self) -> list[str]:
        ordering_fields = ["instrument"]
        for field_name in self.classification_group.get_fields_names(sep="_"):
            ordering_fields.append(f"classification_{field_name}")
        return ordering_fields

    def get_resource_serializer_class(self):
        return {
            "serializer_class_path": "wbfdm.viewsets.instruments.instruments_relationships.get_classified_instrument_serializer_class",
            "serializer_class_method_args": [self.classification_group.id],
        }

    def get_serializer_class(self):
        return get_classified_instrument_serializer_class(self.classification_group.id)

    def get_filterset_class(self, request):
        group = self.classification_group

        attrs = {
            "classification": wb_filters.ModelChoiceFilter(
                label="Height 0",
                queryset=Classification.objects.filter(group=group, height=0),
                endpoint=Classification.get_representation_endpoint(),
                filter_params={"height": 0, "group": group.id},
                value_key=Classification.get_representation_value_key(),
                label_key="{{name}}",
            )
        }
        for index, field_name in enumerate(group.get_fields_names(sep="_"), start=1):
            attrs[f"classification_{field_name}"] = wb_filters.ModelChoiceFilter(
                label=f"Height {index}",
                queryset=Classification.objects.filter(group=group, height=index),
                endpoint=Classification.get_representation_endpoint(),
                filter_params={"height": index, "group": group.id},
                value_key=Classification.get_representation_value_key(),
                method="query_classification",
                label_key="{{name}}",
            )
        filter_class = type("ClassifiedInstrumentFilterSet", (BaseClassifiedInstrumentFilterSet,), attrs)

        def _get_filter_class_for_remote_filter(cls):
            """
            Define which filterset class sender to user for remote filter registration
            """
            return BaseClassifiedInstrumentFilterSet

        filter_class.get_filter_class_for_remote_filter = classmethod(_get_filter_class_for_remote_filter)
        return filter_class

    @cached_property
    def classification_group(self):
        try:
            return ClassificationGroup.objects.get(id=self.request.GET.get("classification_group"))
        except ClassificationGroup.DoesNotExist:
            return get_default_classification_group()

    def get_queryset(self) -> QuerySet[InstrumentClassificationThroughModel]:
        return (
            super()
            .get_queryset()
            .filter(classification__group=self.classification_group)
            .annotate(
                **{
                    f"classification_{field_name}".replace("__", "_"): F(f"classification__{field_name}__name")
                    for field_name in self.classification_group.get_fields_names()
                }
            )
            .select_related(
                *[f"classification__{field_name}" for field_name in self.classification_group.get_fields_names()]
            )
            .prefetch_related(
                "tags",
                Prefetch("instrument", queryset=Instrument.objects.filter(classifications_through__isnull=False)),
            )
        )



from django.conf import settings
from django.utils.module_loading import import_string


def initialize_instruments():
    for sync in map(import_string, getattr(settings, "INSTRUMENT_SYNC", [])):
        sync().update()


def initialize_exchanges():
    for sync in map(import_string, getattr(settings, "EXCHANGE_SYNC", [])):
        sync().update()


def synchronize_instruments():
    for sync in map(import_string, getattr(settings, "INSTRUMENT_SYNC", [])):
        sync().trigger_partial_update()


def synchronize_exchanges():
    for sync in map(import_string, getattr(settings, "EXCHANGE_SYNC", [])):
        sync().trigger_partial_update()



from contextlib import suppress

import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
from wbcore import viewsets
from wbcore.utils.date import get_date_interval_from_request

from wbfdm.analysis.financial_analysis.financial_ratio_analysis import (
    FinancialRatio,
    get_financial_ratios,
)
from wbfdm.filters import FinancialRatioFilterSet
from wbfdm.models.instruments import Instrument

from ..configs import ValuationRatioChartTitleConfig
from ..mixins import InstrumentMixin


class ValuationRatioChartViewSet(InstrumentMixin, viewsets.TimeSeriesChartViewSet):
    queryset = Instrument.objects.all()
    filterset_class = FinancialRatioFilterSet
    title_config_class = ValuationRatioChartTitleConfig

    def get_queryset(self):
        return super().get_queryset().filter(id=self.instrument.id)

    def get_plotly(self, queryset):
        pd.options.plotting.backend = "plotly"

        ratios = [FinancialRatio.PE, FinancialRatio.PS, FinancialRatio.PB, FinancialRatio.PFCF]
        ttm = False if self.request.GET.get("ttm", "true") == "false" else True
        start, end = get_date_interval_from_request(self.request, date_range_fieldname="period")

        df = get_financial_ratios(
            self.instrument.id,
            ratios,
            from_date=start,
            to_date=end,
            ttm=ttm,
        )
        fig = go.Figure()
        colors = iter(px.colors.qualitative.T10)

        for ratio, color in zip(ratios, colors):
            with suppress(AttributeError):
                series = getattr(df, ratio.value)

                fig.add_trace(
                    go.Scatter(
                        x=series.index,
                        y=series,
                        line=dict(color=color),
                        name=f"{ratio.label} {'TTM' if ttm else 'FTM'}",
                        legendgroup=f"{ratio.label} {'TTM' if ttm else 'FTM'}",
                        visible="legendonly" if ratio != FinancialRatio.PE else True,
                    ),
                )

        # fig = df.plot()
        fig.update_layout(
            template="plotly_white",
            legend=dict(x=0.02, y=0.98, bgcolor="rgba(0,0,0,0)"),
            margin=dict(l=0, r=0, t=0, b=0),
            showlegend=True,
        )

        for i, d in enumerate(fig.data):
            with suppress(AttributeError, IndexError, TypeError, OverflowError):
                if (v := d.y[-1]) and not np.isnan(v):
                    fig.add_scatter(
                        x=[d.x[-1]],
                        y=[v],
                        name=d.name,
                        mode="markers+text",
                        text=round(v),
                        textfont=dict(color=d.line.color),
                        textposition="middle right",
                        marker=dict(color=d.line.color, size=12, symbol="circle"),
                        legendgroup=d.name,
                        showlegend=False,
                        visible="legendonly" if i != 0 else True,
                    )

        return fig



from rest_framework import serializers
from rest_framework.reverse import reverse
from wbcore import serializers as wb_serializers
from wbcore.contrib.directory.models import Person
from wbcore.contrib.directory.serializers import PersonRepresentationSerializer
from wbcore.contrib.tags.serializers import TagSerializerMixin

from wbfdm.models import (
    Classification,
    Instrument,
    InstrumentClassificationRelatedInstrument,
    InstrumentClassificationThroughModel,
    InstrumentFavoriteGroup,
    RelatedInstrumentThroughModel,
)

from .classifications import ClassificationZeroHeightRepresentationSerializer
from .instruments import (
    ClassifiableInstrumentRepresentationSerializer,
    InvestableInstrumentRepresentationSerializer,
    PrimaryInvestableInstrumentRepresentationSerializer,
)


class InstrumentFavoriteGroupRepresentationSerializer(wb_serializers.RepresentationSerializer):
    _detail = wb_serializers.HyperlinkField(reverse_name="wbfdm:favoritegroup-detail")

    class Meta:
        model = InstrumentFavoriteGroup
        fields = ("id", "name", "owner", "public", "_detail")


class InstrumentFavoriteGroupModelSerializer(wb_serializers.ModelSerializer):
    _instruments = PrimaryInvestableInstrumentRepresentationSerializer(source="instruments", many=True)
    owner = wb_serializers.PrimaryKeyRelatedField(
        queryset=lambda: Person.objects.filter_only_internal(),
        default=wb_serializers.CurrentUserDefault("profile"),
    )
    _owner = PersonRepresentationSerializer(source="owner")

    def validate(self, data):
        if not data.get("owner", None):
            data["owner"] = self.context["request"].user.profile if self.context.get("request") else None
        return data

    class Meta:
        model = InstrumentFavoriteGroup
        fields = (
            "id",
            "name",
            "owner",
            "_owner",
            "public",
            "primary",
            "instruments",
            "_instruments",
            "_additional_resources",
        )


class ReadOnlyInstrumentFavoriteGroupModelSerializer(InstrumentFavoriteGroupModelSerializer):
    class Meta(InstrumentFavoriteGroupModelSerializer.Meta):
        read_only_fields = InstrumentFavoriteGroupModelSerializer.Meta.fields


class RelatedInstrumentThroughInstrumentModelSerializer(wb_serializers.ModelSerializer):
    _related_instrument = InvestableInstrumentRepresentationSerializer(source="related_instrument")
    _instrument = InvestableInstrumentRepresentationSerializer(source="instrument")

    class Meta:
        model = RelatedInstrumentThroughModel
        fields = (
            "id",
            "_related_instrument",
            "related_instrument",
            "_instrument",
            "instrument",
            "related_type",
            "is_primary",
        )


class InstrumentClassificationThroughModelSerializer(TagSerializerMixin, wb_serializers.ModelSerializer):
    instrument = wb_serializers.PrimaryKeyRelatedField(
        queryset=Instrument.objects.all(), default=wb_serializers.DefaultFromKwargs("instrument_id")
    )
    _instrument = ClassifiableInstrumentRepresentationSerializer(source="instrument")
    classification = wb_serializers.PrimaryKeyRelatedField(
        queryset=Classification.objects.all(), default=wb_serializers.DefaultFromKwargs("classification_id")
    )
    _classification = ClassificationZeroHeightRepresentationSerializer(source="classification")
    reason = wb_serializers.TextAreaField(label="Reason", allow_null=True, allow_blank=True, required=False)
    _related_instruments = ClassifiableInstrumentRepresentationSerializer(source="related_instruments", many=True)

    @wb_serializers.register_resource()
    def related_instrument_list(self, instance, request, user):
        return {"related_instruments": reverse("wbfdm:related_instrument-list", args=[instance.id], request=request)}

    def validate(self, data):
        instrument = data.get("instrument", self.instance.instrument if self.instance else None)
        classification = data.get("classification", self.instance.classification if self.instance else None)
        if not instrument:
            raise serializers.ValidationError({"instrument": "Instrument cannot be null"})
        if not classification:
            raise serializers.ValidationError({"classification": "Classification cannot be null"})

        if (
            not self.instance
            and InstrumentClassificationThroughModel.objects.filter(
                classification=classification, instrument=instrument
            ).exists()
        ):
            if hasattr(data, "instrument"):
                raise serializers.ValidationError(
                    {
                        "instrument": f"A relationship already exists between {instrument} and {classification}",
                    }
                )
            else:
                raise serializers.ValidationError(
                    {
                        "classification": f"A relationship already exists between {instrument} and {classification}",
                    }
                )

        return data

    class Meta:
        model = InstrumentClassificationThroughModel
        fields = (
            "id",
            "instrument",
            "classification",
            "_instrument",
            "_classification",
            "related_instruments",
            "_related_instruments",
            "is_favorite",
            "reason",
            "pure_player",
            "top_player",
            "percent_of_revenue",
            "tags",
            "_tags",
            "_additional_resources",
        )
        percent_fields = ["percent_of_revenue"]


class InstrumentClassificationRelatedInstrumentRepresentationSerializer(wb_serializers.ModelSerializer):
    class Meta:
        model = InstrumentClassificationRelatedInstrument
        fields = ("id", "related_instrument")


class InstrumentClassificationRelatedInstrumentModelSerializer(wb_serializers.ModelSerializer):
    _related_instrument = ClassifiableInstrumentRepresentationSerializer(source="related_instrument")

    class Meta:
        model = InstrumentClassificationRelatedInstrument
        fields = (
            "id",
            "classified_instrument",
            # "_classified_instrument",
            "related_instrument",
            "_related_instrument",
            "related_instrument_type",
        )



# Generated by Django 5.0.8 on 2024-09-25 09:37

from django.db import migrations, models
from wbcore.contrib.currency.models import Currency
from wbfdm.models import Instrument, InstrumentType


def create_cash_equivalents(apps, schema_editor):
    for currency in Currency.objects.all():
        Instrument.objects.create(
            name=f"Cash Equivalent {currency.key}",
            instrument_type=InstrumentType.CASHEQUIVALENT,
            currency=currency,
            is_cash_equivalent=True,
        )


class Migration(migrations.Migration):
    dependencies = [
        ("wbfdm", "0025_instrument_is_primary_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="instrument",
            name="is_cash_equivalent",
            field=models.BooleanField(default=False),
        ),
        migrations.RunPython(create_cash_equivalents),
    ]



from datetime import datetime
from io import BytesIO
from typing import Optional

import pandas as pd
from django.db import models
from pandas.tseries.offsets import BDay
from wbcore.contrib.io.backends import AbstractDataBackend, register

from .mixin import DataBackendMixin
from .utils import Controller

DEFAULT_MAPPING = {
    "MV": "market_capitalization",
    "MVC": "market_capitalization_consolidated",
    "VO": "volume",
    "MAV#(X(VO),-50D,R)": "volume_50d",
    "BETA": "beta",
    "NOSH": "outstanding_shares",
    "IBNOSH": "outstanding_shares_consolidated",
    "P": "close",
    "PI": "price_index",
    "RY": "yield_redemption",
    "NR": "net_return",
    "IO": "offered_rate",
}

MUTUAL_FUND_TIMEDELTA_DAY_SHIFT = 7


@register("Instrument Price", provider_key="refinitiv", save_data_in_import_source=False, passive_only=False)
class DataBackend(DataBackendMixin, AbstractDataBackend):
    def __init__(self, import_credential: Optional[models.Model] = None, **kwargs):
        self.controller = Controller(import_credential.username, import_credential.password)

    def get_files(
        self,
        execution_time: datetime,
        obj_external_ids: list[str] = None,
        **kwargs,
    ) -> BytesIO:
        execution_date = execution_time.date()

        start = kwargs.get("start", (execution_date - BDay(2)).date())

        delayed_instrument_ids = list(
            filter(
                lambda x: True,
                obj_external_ids,
            )
        )
        obj_external_ids = list(filter(lambda v: v not in delayed_instrument_ids, obj_external_ids))
        # we get all active instruments even these were we are not suppose to fetch prices.
        fields = list(DEFAULT_MAPPING.keys())
        if obj_external_ids or delayed_instrument_ids:
            df_list = []
            if obj_external_ids:
                df_list.append(
                    self.controller.get_data(obj_external_ids, fields, start, execution_date, ibes_fields=["IBNOSH"])
                )
            if delayed_instrument_ids:
                # we need to get mutual fund price in a different batch because there is a price delay and the windows approach can't handle it otherwise
                df_list.append(
                    self.controller.get_data(
                        delayed_instrument_ids,
                        fields,
                        start - BDay(MUTUAL_FUND_TIMEDELTA_DAY_SHIFT),
                        execution_date,
                        ibes_fields=["IBNOSH"],
                    )
                )
            df = pd.concat(df_list, axis=0, ignore_index=True)
            if not df.empty:
                content_file = BytesIO()
                df.to_json(content_file, orient="records")
                file_name = f"instrument_price_{start:%Y-%m-%d}-{execution_date:%Y-%m-%d}_{datetime.timestamp(execution_time)}.json"
                yield file_name, content_file



# Generated by Django 4.2.7 on 2023-11-27 09:12

import django.db.models.deletion
import mptt
from django.apps import apps
from django.db import migrations, models


def missing_operations():
    if apps.is_installed("wbportfolio"):
        return [
            migrations.AddField(
                model_name="instrument",
                name="source",
                field=models.CharField(blank=True, max_length=64, null=True),
            ),
            migrations.AddField(
                model_name="instrument",
                name="source_id",
                field=models.CharField(blank=True, max_length=64, null=True),
            ),
            migrations.AddField(
                model_name="exchange",
                name="source",
                field=models.CharField(blank=True, max_length=64, null=True),
            ),
            migrations.AddField(
                model_name="exchange",
                name="source_id",
                field=models.CharField(blank=True, max_length=64, null=True),
            ),
        ]
    return []


class Migration(migrations.Migration):
    dependencies = (
        [
            ("wbfdm", "0010_classification_classificationgroup_deal_exchange_and_more"),
            ("wbportfolio", "0052_remove_cash_instrument_ptr_and_more"),
        ]
        if apps.is_installed("wbportfolio")
        else [
            ("wbfdm", "0010_classification_classificationgroup_deal_exchange_and_more"),
        ]
    )

    operations = [
        migrations.AddField(
            model_name="instrument",
            name="corporate_actions",
            field=models.JSONField(default=dict),
        ),
        migrations.AddField(
            model_name="instrument",
            name="dividends",
            field=models.JSONField(default=dict),
        ),
        migrations.AddField(
            model_name="instrument",
            name="adjustments",
            field=models.JSONField(default=dict),
        ),
        migrations.AddField(
            model_name="instrument",
            name="officers",
            field=models.JSONField(default=dict),
        ),
        migrations.AddField(
            model_name="instrument",
            name="esg",
            field=models.JSONField(default=dict),
        ),
        migrations.AddField(
            model_name="instrument",
            name="esg_controversies",
            field=models.JSONField(default=dict),
        ),
        migrations.AddField(
            model_name="instrument",
            name="financials",
            field=models.JSONField(default=dict),
        ),
        migrations.AddField(
            model_name="instrument",
            name="level",
            field=models.PositiveIntegerField(default=0, editable=False),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="instrument",
            name="lft",
            field=models.PositiveIntegerField(default=0, editable=False),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="instrument",
            name="market_data",
            field=models.JSONField(default=dict),
        ),
        migrations.AddField(
            model_name="instrument",
            name="parent",
            field=mptt.fields.TreeForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="children",
                to="wbfdm.instrument",
                verbose_name="Parent Instrument",
            ),
        ),
        migrations.AddField(
            model_name="instrument",
            name="rght",
            field=models.PositiveIntegerField(default=0, editable=False),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="instrument",
            name="statements",
            field=models.JSONField(default=dict),
        ),
        migrations.AddField(
            model_name="instrument",
            name="tree_id",
            field=models.PositiveIntegerField(db_index=True, default=0, editable=False),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name="instrument",
            name="description",
            field=models.TextField(blank=True, default=""),
        ),
        migrations.AlterField(
            model_name="instrument",
            name="instrument_type",
            field=models.CharField(
                blank=True,
                choices=[
                    ("ADR", "American Depository Receipt (ADR)"),
                    ("CF", "Close Ended Fund"),
                    ("EQ", "Equity"),
                    ("ETF", "Exchange Traded Fund"),
                    ("ETC", "Exchange Traded Commodity"),
                    ("ETN", "Exchange Traded Note"),
                    ("WARRANT", "Warrant"),
                    ("GDR", "Global Depository Receipt (GDR)"),
                    ("GS", "Genussschein"),
                    ("TRUST", "Investment Trust"),
                    ("NVDR", "Non-voting depository receipt (NVDR)"),
                    ("PREF", "Preference Share"),
                    ("COMP", "Company"),
                    ("QT", "Quote"),
                    ("Product", "Product"),
                    ("Index", "Index"),
                    ("ProductGroup", "Product Group"),
                    ("Cash", "Cash"),
                ],
                max_length=32,
                null=True,
            ),
        ),
        *missing_operations(),
        migrations.AddConstraint(
            model_name="instrument",
            constraint=models.UniqueConstraint(fields=("source_id", "source"), name="unique_source"),
        ),
    ]



from contextlib import suppress

import pandas as pd


def normalize_series(series: pd.Series, base_series: pd.Series | None = None) -> pd.Series:
    """
    Normalize the given time series by aligning and adjusting its values based on a base series.

    Args:
        series (pd.Series): The time series to be normalized.
        base_series (pd.Series | None): The base time series for alignment and normalization.
            If None, no alignment or normalization is performed.

    Returns:
        pd.Series: The normalized time series.

    Note:
        - If base_series is provided, the function aligns the series with the base_series and normalizes
          the values based on the first value of both series.
        - If the series index is older than the base series index, the function trims the series accordingly.
        - If the first value of the series is zero or if the base series first value is zero, no normalization
          is performed.
    """
    if base_series is not None and not base_series.empty:
        with suppress(KeyError, IndexError):
            # Ensure the series is not older than the the base serie
            series = series.loc[series.index >= base_series.index.min()]
            # if a base serie is provided and normalization is possible, we align the timeseries
            if series.iloc[0] != 0 and (normalize_factor := base_series.iloc[0]):
                series = series / series.iloc[0] * normalize_factor
    return series



from datetime import datetime
from io import BytesIO

from wbcore.contrib.io.backends import register

from .fundamental import DataBackend as ParentDataBackend

DEFAULT_MAPPING = {
    "WC19601": "segment_1_sales",
    "WC19600 ": "segment_1_description",
    "WC19611": "segment_2_sales",
    "WC19610 ": "segment_2_description",
    "WC19621": "segment_3_sales",
    "WC19620 ": "segment_3_description",
}


@register("Geographic Segment", provider_key="refinitiv", save_data_in_import_source=False, passive_only=False)
class DataBackend(ParentDataBackend):
    def get_files(
        self,
        execution_time: datetime,
        obj_external_ids: list[str] = None,
        **kwargs,
    ) -> BytesIO:
        yield from super().get_files(
            execution_time,
            obj_external_ids=obj_external_ids,
            fields=list(DEFAULT_MAPPING.keys()),
            filename="geographic_segment",
            **kwargs,
        )



from datetime import date

from pandas.tseries.offsets import BDay, BMonthBegin, BMonthEnd
from psycopg.types.range import DateRange
from wbcore.utils.date import current_quarter_date_start

from wbfdm.models import Classification, Instrument
from wbfdm.preferences import get_default_classification_group


def _get_default_classification_group_id(field, request, view, **kwargs):
    if classification_id := view.kwargs.get("classification_id", None):
        return Classification.objects.get(id=classification_id).group.id
    if group := get_default_classification_group():
        return group.id
    return None


def get_earliest_date(field, request, view):
    instrument_id = None
    if "instrument_id" in view.kwargs:
        instrument_id = view.kwargs["instrument_id"]
    if "instrument_id" in request.GET:
        instrument_id = request.GET["instrument_id"]
    if instrument_id:
        _date = Instrument.objects.get(id=instrument_id).inception_date
    else:
        _date = current_quarter_date_start()
    return (_date + BDay(0)).date()


def get_latest_date(field, request, view, **kwargs):
    instrument_id = None
    if "instrument_id" in view.kwargs:
        instrument_id = view.kwargs["instrument_id"]
    if "instrument_id" in request.GET:
        instrument_id = request.GET["instrument_id"]
    if instrument_id and (instrument := Instrument.objects.get(id=instrument_id)) and instrument.delisted_date:
        _date = instrument.delisted_date
    else:
        _date = date.today()
    return (_date - BDay(0)).date()


def last_period_start(*args, **kwargs):
    return (date.today() - BMonthBegin(1)).date()


def last_period_end(*args, **kwargs):
    return (date.today() + BMonthEnd(1)).date()


def last_period_date_range(*args, **kwargs):
    return DateRange(last_period_start(*args, **kwargs), last_period_end(*args, **kwargs))



from .exchanges import ExchangeMenuItem
from .classifications import ClassificationMenuItem, ClassificationGroupMenuItem
from .instrument_lists import INSTRUMENT_LIST_MENUITEM
from .instruments import INSTRUMENT_MENUITEM, INVESTABLE_UNIVERSE_MENUITEM
from .instruments_relationships import (
    INSTRUMENTFAVORITEGROUP_MENUITEM,
    INSTRUMENT_REQUEST_MENUITEM,
    CLASSIFIEDINSTRUMENTS_MENUITEM,
)



# Generated by Django 5.0.9 on 2024-10-11 08:34

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("currency", "0001_initial"),
        ("geography", "0001_initial"),
        ("io", "0007_alter_exportsource_query_params"),
        ("tags", "0001_initial"),
        ("wbfdm", "0026_instrument_is_cash_equivalent"),
    ]

    operations = [
        migrations.RemoveConstraint(
            model_name="instrument",
            name="unique_ric",
        ),
        migrations.RemoveConstraint(
            model_name="instrument",
            name="unique_rmc",
        ),
        migrations.RemoveConstraint(
            model_name="instrument",
            name="unique_isin",
        ),
        migrations.RemoveConstraint(
            model_name="instrument",
            name="unique_sedol",
        ),
        migrations.RemoveConstraint(
            model_name="instrument",
            name="unique_valoren",
        ),
        migrations.RemoveConstraint(
            model_name="instrument",
            name="unique_cusip",
        ),
        migrations.RemoveConstraint(
            model_name="instrument",
            name="unique_instrument_primary",
        ),
        migrations.AddConstraint(
            model_name="instrument",
            constraint=models.UniqueConstraint(
                condition=models.Q(("is_security", True), ("delisted_date__isnull", True)),
                fields=("refinitiv_identifier_code",),
                name="unique_ric",
            ),
        ),
        migrations.AddConstraint(
            model_name="instrument",
            constraint=models.UniqueConstraint(
                condition=models.Q(("is_security", True), ("delisted_date__isnull", True)),
                fields=("refinitiv_mnemonic_code",),
                name="unique_rmc",
            ),
        ),
        migrations.AddConstraint(
            model_name="instrument",
            constraint=models.UniqueConstraint(
                condition=models.Q(("is_security", True), ("delisted_date__isnull", True)),
                fields=("isin",),
                name="unique_isin",
            ),
        ),
        migrations.AddConstraint(
            model_name="instrument",
            constraint=models.UniqueConstraint(
                condition=models.Q(("is_security", True), ("delisted_date__isnull", True)),
                fields=("sedol",),
                name="unique_sedol",
            ),
        ),
        migrations.AddConstraint(
            model_name="instrument",
            constraint=models.UniqueConstraint(
                condition=models.Q(("is_security", True), ("delisted_date__isnull", True)),
                fields=("valoren",),
                name="unique_valoren",
            ),
        ),
        migrations.AddConstraint(
            model_name="instrument",
            constraint=models.UniqueConstraint(
                condition=models.Q(("is_security", True), ("delisted_date__isnull", True)),
                fields=("cusip",),
                name="unique_cusip",
            ),
        ),
        migrations.AddConstraint(
            model_name="instrument",
            constraint=models.UniqueConstraint(
                condition=models.Q(("is_primary", True), ("is_managed", False)),
                fields=("parent", "is_primary"),
                name="unique_instrument_primary",
            ),
        ),
    ]



from rest_framework.reverse import reverse
from wbcore.metadata.configs.endpoints import EndpointViewConfig


class InstrumentESGPAIEndpointViewConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return reverse(
            "wbfdm:pai-list",
            args=[self.view.kwargs["instrument_id"]],
            request=self.request,
        )


class InstrumentESGControversiesEndpointViewConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return reverse(
            "wbfdm:controversies-list",
            args=[self.view.kwargs["instrument_id"]],
            request=self.request,
        )



from datetime import date
from math import pow

import numpy as np
import pandas as pd
from dateutil.relativedelta import relativedelta


class FinancialStatistics:
    def __init__(self, prices: pd.Series):
        prices = prices.copy().astype("float")
        if type(prices) is not pd.Series:
            raise ValueError("prices is expected to be a pandas Series")
        self.prices = prices.sort_index()
        if not prices.empty:
            self.start = self.prices.index[0]
            self.end = self.prices.index[-1]
        else:
            self.start = self.end = None
        self.prices.index = pd.to_datetime(self.prices.index)

    def is_valid(self):
        return not self.prices.empty

    def extract_monthly_performance_df(self) -> pd.Series:
        return self.compute_performance()

    def extract_annual_performance_df(self) -> pd.Series:
        return self.compute_performance(freq="BYE")

    def extract_inception_performance_df(self) -> float:
        if not self.prices.empty and self.prices.iloc[0]:
            return self.prices.iloc[-1] / self.prices.iloc[0] - 1
        return 0

    def extract_daily_performance_df(self) -> pd.Series:
        return self.compute_performance(freq="B")

    # ---------------------------- Financial Statistics ----------------------------
    # ------------------------------------------------------------------------------

    def compute_performance(self, freq: str | int = "BME", extrapolate: bool = True) -> pd.Series:
        perfs = self.prices
        if not perfs.empty:
            if isinstance(freq, int):
                perfs = perfs.resample(f"{freq}D", origin="end").ffill().pct_change()
            else:
                perfs = perfs.asfreq(freq, method="ffill")
                if extrapolate:
                    perfs.loc[self.prices.index[0]] = self.prices.iloc[0]
                perfs.loc[self.prices.index[-1]] = self.prices.iloc[-1]
                perfs = perfs.sort_index()
                perfs = perfs / perfs.shift(1) - 1
        return perfs.dropna(how="any").rename("performance")

    def get_mean_return(self, freq: str = "B") -> float | None:
        """Calculates the mean of returns"""
        if self.is_valid():
            return self.compute_performance(freq=freq).mean()
        return None

    def get_best_and_worst_returns(self, top: int = 10, freq: str = "B") -> pd.DataFrame:
        """Returns a DataFrame with the top best returns and top worst returns"""
        df = pd.DataFrame(columns=["Date Best Return", "Best Return", "Date Worst Return", "Worst Return"])
        if self.is_valid():
            sort_returns = self.compute_performance(freq=freq).sort_values()
            df.loc[:, "Worst Return"] = sort_returns[0:top].values
            df.loc[:, "Date Worst Return"] = sort_returns[0:top].index
            sort_returns = sort_returns.sort_values(ascending=False)
            df.loc[:, "Best Return"] = sort_returns[0:top].values
            df.loc[:, "Date Best Return"] = sort_returns[0:top].index
            df["Date Worst Return"] = pd.to_datetime(df["Date Worst Return"])
            df["Date Best Return"] = pd.to_datetime(df["Date Best Return"])
            return df
        return df

    def get_drawdowns(self, freq: str = "B") -> pd.Series:
        """
        Investopedia: A drawdown is a peak-to-trough decline during a specific period for an investment,
        trading account, or fund. A drawdown is usually quoted as the percentage between the peak and the
        subsequent trough.
        """
        high_net_value = self.prices.cummax()  # keep highest price the column
        drawdowns = self.prices / high_net_value - 1
        if freq:
            drawdowns = drawdowns.asfreq(freq, method="ffill")
        return drawdowns

    def get_last_drawdowns(self) -> pd.Series:
        """Returns a Dataframe with the last drawdowns"""
        df = self.get_drawdowns()
        last_highest_price_date = df[df == 0].index[-1].strftime("%Y-%m-%d")
        return df.loc[last_highest_price_date : df.index[-1]]

    def get_maximum_drawdown(self) -> float:
        """Returns the maximum drawdown."""
        df = self.get_drawdowns()
        maximum_dd = df.min()  # maximum in negative value
        return maximum_dd

    def get_last_recent_maximum_drawdown(self) -> float:
        """Returns the last recent maximum drawdown."""
        df = self.get_last_drawdowns()
        maximum_dd = df.min()  # maximum in negative value
        return maximum_dd

    def get_maximum_drawdown_date(self) -> float:
        df = self.get_drawdowns()
        mdd_date = df[df == df.min()].index.to_pydatetime()[0].strftime("%Y-%m-%d")
        return mdd_date

    def get_last_recent_maximum_drawdown_date(self) -> date:
        """Returns the last recent maximum drawdown date."""
        df = self.get_last_drawdowns()
        mdd_date = df[df == df.min()].index.to_pydatetime()[0].strftime("%Y-%m-%d")
        return mdd_date

    def get_longest_drawdown_period(self) -> int:
        """Returns the longest number of day during drawdown"""
        df = self.get_drawdowns()
        df["Drawdown Period"] = df.mask(df < 0.0, 1)
        a = df["Drawdown Period"] != 0
        df1 = a.cumsum() - a.cumsum().where(~a).ffill().fillna(0).astype(int)
        return df1.max()

    def get_cumulative_returns(self, freq: str = "B") -> pd.Series:
        """Returns a dataframe with the cumulative return each day."""
        returns = self.compute_performance(freq=freq)
        # R = [(1+r1) * (1 + r2) * ... * (1+rn)] - 1
        return returns.add(1).cumprod() - 1

    def get_last_cumulative_return(self, freq: str = "B") -> float:
        """Returns the last cumulative return."""
        cumulative_returns = self.get_cumulative_returns(freq=freq)
        return cumulative_returns.iat[-1]

    def get_positive_outliers(self, quantile: float = 0.95, freq: str = "B") -> pd.Series:
        """Extracts the outliers in a DataFrame."""
        returns = self.compute_performance(freq=freq)
        return returns[returns > returns.quantile(quantile)].dropna(how="all")

    def get_value_at_risk(self, alpha: float = 0.05, freq: str = "B") -> float:
        """Calculates the value at risk for a given confident interval."""
        if alpha < 0 or alpha > 1:
            raise ValueError(f"Alpha = {alpha} is not possible, alpha must be between 0 and 1.")

        returns = self.compute_performance(freq=freq)
        return returns.quantile(alpha, interpolation="higher")

    def get_conditional_value_at_risk(self, alpha: float = 0.05, freq: str = "B") -> float:
        """Calculates the conditional value at risk for a given confident interval."""
        if alpha < 0 or alpha > 1:
            raise ValueError(f"Alpha = {alpha} is not possible, alpha must be between 0 and 1.")

        returns = self.compute_performance(freq=freq)
        return returns[returns <= returns.quantile(alpha, interpolation="lower")].dropna(how="all").mean()

    def get_skewness(self, freq: str = "B") -> float:
        """Calculates the skewness of the returns."""
        returns = self.compute_performance(freq=freq)
        return returns.skew()

    def get_kurtosis(self, freq: str = "B") -> float:
        """Calculates the Kurtosis of the returns."""
        returns = self.compute_performance(freq=freq)
        return returns.kurtosis()

    def get_excess_kurtosis(self, freq: str = "B") -> float:
        return self.get_kurtosis(freq=freq) - 3.0

    def get_compound_annual_growth_rate(self) -> float:
        """Calculates the compound annual growth rate (Last Price / First Price)(¹/year) - 1"""
        first_price, last_price = self.prices.iat[0], self.prices.iat[-1]
        years = (self.prices.index[-1] - self.prices.index[0]).days / 365.0
        return (last_price / first_price) ** (1 / years) - 1 if first_price != 0 and years != 0 else 0

    def get_sortino_ratio(self, target_ratio: float = 0, freq: str = "B") -> float:
        """
        Calculates the adjusted Sortino ratio:
        Sortino ratio = (Rp - target_ratio) / below_target_semi_deviation
        below_target_semi_deviation = √(1 / T * Σ (d²))
        d = (Rt-1 - target_ratio) only if the result is negative
        """
        returns = self.compute_performance(freq=freq)
        below_target_sd = np.sqrt((returns[returns < target_ratio] ** 2).sum() / len(returns))
        annualized_return = self.get_compound_annual_growth_rate()
        below_target_annualized = below_target_sd * np.sqrt(260)
        return (annualized_return - target_ratio) / below_target_annualized

    def get_adjusted_sortino_ratio(self, freq: str = "B") -> float:
        """Calculates the adjusted Sortino ratio (Sortino Ratio / √(2))"""
        return self.get_sortino_ratio(freq=freq) / np.sqrt(2)

    def get_calmar_ratio(self) -> float:
        """Calculates the calmar ratio (CAGR% / MaxDD%)"""
        compound_annual_growth_rate = self.get_compound_annual_growth_rate()
        maximum_dd = self.get_maximum_drawdown()
        return compound_annual_growth_rate / abs(maximum_dd)

    def get_risk_free_rate(self, risk_free_rate_prices_df: pd.Series) -> float | None:
        # if the instrument has no prices, we do not know the lifetime of instrument, so we do not know for the risk
        # instrument too
        if not risk_free_rate_prices_df.empty:
            nb_days = (self.prices.index[-1] - self.prices.index[0]).days
            nb_days = 1 if nb_days == 0 else nb_days
            exp = 365.0 / nb_days
            annualized_free_rate = 1 + risk_free_rate_prices_df.sort_index(ascending=False) / (365 * 100)
            tmp = annualized_free_rate.shift(1, fill_value=1).cumprod().dropna()
            risk_free_rate = pow(tmp.iloc[-1], exp) - 1
            return risk_free_rate

    def get_sterling_ratio(self, risk_free_rate_prices_df: pd.Series) -> float | None:
        """Calculates the adjusted Sterling ratio ((Rp- Rf) / MDD)"""
        if not risk_free_rate_prices_df.empty:
            maximum_dd = self.get_maximum_drawdown()
            annualized_return = self.get_compound_annual_growth_rate()
            rf = self.get_risk_free_rate(risk_free_rate_prices_df)
            rf = 0 if not rf else rf
            return (annualized_return - rf) / abs(maximum_dd)

    def get_burke_ratio(self, risk_free_rate_prices_df: pd.Series) -> float | None:
        """Calculates the adjusted Burke ratio ((Rp- Rf) / √(Σ (DD²)))"""
        if not risk_free_rate_prices_df.empty:
            drawdowns = self.get_drawdowns()
            drawdowns = np.sqrt(drawdowns[drawdowns != 0].pow(2).sum())
            annualized_return = self.get_compound_annual_growth_rate()
            rf = self.get_risk_free_rate(risk_free_rate_prices_df)
            rf = 0 if not rf else rf
            return (annualized_return - rf) / drawdowns

    def get_volatility(self, freq: str = "B") -> float:
        """Calculates the volatility of returns"""
        adj = 1
        if freq == "B":
            adj = 260
        elif freq == "W-MON":
            adj = 52
        elif freq == "BME":
            adj = 12
        return self.compute_performance(freq=freq).std() * np.sqrt(adj)

    def get_annualized_daily_volatility(self) -> float:
        """Computed the annualized volatility"""
        df = self.compute_performance(freq=1)
        last_date = df.index[-1]
        last_year = last_date - relativedelta(months=12)
        df = df.loc[df.index >= last_year]
        return df.std()

    def prepare_daily_returns_instrument_vs_benchmark(self, benchmark_prices_df: pd.Series) -> pd.DataFrame:
        if benchmark_prices_df.empty:
            raise ValueError("benchmark price cannot be empty")
        df = pd.DataFrame()
        df["benchmark"] = FinancialStatistics(benchmark_prices_df).extract_daily_performance_df()
        df["instrument"] = self.extract_daily_performance_df()
        return df

    def get_beta(self, benchmark_prices_df: pd.Series) -> float | None:
        if not benchmark_prices_df.empty:
            df = self.prepare_daily_returns_instrument_vs_benchmark(benchmark_prices_df)
            cov = df["instrument"].cov(df["benchmark"])
            var = df["benchmark"].var()
            beta = cov / var
            return beta

    def get_correlation(self, benchmark_prices_df: pd.Series) -> float | None:
        if not benchmark_prices_df.empty:
            df = self.prepare_daily_returns_instrument_vs_benchmark(benchmark_prices_df)
            return df["instrument"].corr(df["benchmark"])

    def get_sharpe_ratio(self, risk_free_rate_prices_df: pd.Series) -> float | None:
        if not risk_free_rate_prices_df.empty:
            annualized_volatility = self.get_volatility()
            annualized_return = self.get_compound_annual_growth_rate()
            rf = self.get_risk_free_rate(risk_free_rate_prices_df)
            rf = 0 if not rf else rf
            return (annualized_return - rf) / annualized_volatility

    # -------------------------- End of Financial Statistics -----------------------
    # ------------------------------------------------------------------------------



# Generated by Django 4.2.8 on 2023-12-22 07:36

from contextlib import suppress

import django.db.models.deletion
from django.apps import apps as django_apps
from django.contrib.auth.models import Group, Permission
from django.contrib.contenttypes.management import create_contenttypes
from django.contrib.contenttypes.models import ContentType
from django.db import migrations, models
from django.utils import timezone
from wbcore.contrib.io.models import ImportedObjectProviderRelationship

InstrumentTypeMap = {
    "ADR": ("american_depository_receipt", "American Depository Receipt (ADR)"),
    "CF": ("close_ended_fund", "Close Ended Fund"),
    "MutualFund": ("mutual_fund", "Mutual Fund"),
    "EQ": ("equity", "Equity"),
    "Equity": ("equity", "Equity"),
    "PrivateEquity": ("private_equity", "Private Equity"),
    "ETF": ("exchange_traded_fund", "Exchange Traded Fund"),
    "ETC": ("exchange_traded_commodity", "Exchange Traded Commodity"),
    "ETP": ("exchange_traded_product", "Exchange Traded Product"),
    "ETN": ("exchange_traded_note", "Exchange Traded Note"),
    "WARRANT": ("warrant", "Warrant"),
    "GDR": ("global_depository_receipt", "Global Depository Receipt (GDR)"),
    "GS": ("genussschein", "Genussschein"),
    "TRUST": ("investment_trust", "Investment Trust"),
    "NVDR": ("non_voting_depository_receipt", "Non-voting depository receipt (NVDR)"),
    "PREF": ("preference_share", "Preference Share"),
    "COMP": ("company", "Company"),
    "QT": ("quote", "Quote"),
    "Product": ("product", "Product"),
    "Index": ("index", "Index"),
    "ProductGroup": ("product_group", "Product Group"),
    "Cash": ("cash", "Cash"),
    "InterestRateDerivative": ("interest_rate_derivative", "Interest Rate Derivative"),
}


def migrate_instrument_type(apps, schema_editor):
    Instrument = apps.get_model("wbfdm", "Instrument")
    InstrumentType = apps.get_model("wbfdm", "InstrumentType")
    for old_instrument_type, (instrument_type_key, instrument_type_label) in InstrumentTypeMap.items():
        label_splits = instrument_type_label.split(" ")
        short_name = (
            "".join(word[0].upper() for word in label_splits) if len(label_splits) > 1 else instrument_type_label
        )
        instrument_type = InstrumentType.objects.get_or_create(
            key=instrument_type_key, defaults={"name": instrument_type_label, "short_name": short_name}
        )[0]
        Instrument.objects.filter(old_instrument_type=old_instrument_type).update(instrument_type=instrument_type)


def migrate_bbg_code(apps, _schema_editor):
    Exchange = apps.get_model("wbfdm", "Exchange")
    for exchange in Exchange.objects.all():
        exchange.bbg_exchange_codes = [exchange.bbg_exchange]
        exchange.save()


def post_migrate_script(apps, schema_editor):
    from wbcore.contrib.authentication.models import User

    Instrument = apps.get_model("wbfdm", "Instrument")

    # apps.get_model("wbfdm", "InstrumentPrice").objects.update(created=models.F("date"))
    # ensure content type are created at this point
    create_contenttypes(django_apps.get_app_config("wbfdm"))

    # move old permission to new permission
    for old_permission in Permission.objects.filter(content_type__app_label="wbportfolio"):
        with suppress(Permission.DoesNotExist, ContentType.DoesNotExist):
            new_permission = Permission.objects.get(
                codename=old_permission.codename.replace("wbportfolio", "wbfdm"),
                content_type=ContentType.objects.get(app_label="wbfdm", model=old_permission.content_type.model),
            )
            for user in User.objects.filter(user_permissions=old_permission):
                user.user_permissions.remove(old_permission)
                user.user_permissions.add(new_permission)
            for group in Group.objects.filter(permissions=old_permission):
                group.permissions.remove(old_permission)
                group.permissions.add(new_permission)

    for cb_insight_rel in ImportedObjectProviderRelationship.objects.filter(
        content_type__model="equity", provider__key="cbinsights"
    ):
        with suppress(Instrument.DoesNotExist):
            instrument = Instrument.objects.get(id=cb_insight_rel.object_id)
            instrument.source = "cbinsights"
            instrument.source_id = cb_insight_rel.provider_identifier
            instrument.save()

    for refinitiv_insight_rel in ImportedObjectProviderRelationship.objects.filter(
        content_type__model="instrument", provider__key="refinitiv"
    ):
        with suppress(Instrument.DoesNotExist):
            instrument = Instrument.objects.get(id=refinitiv_insight_rel.object_id)
            instrument.source = "refinitiv"
            instrument.source_id = refinitiv_insight_rel.provider_identifier
            instrument.save()

    InstrumentLookup = apps.get_model("wbfdm", "InstrumentLookup")
    for rel in InstrumentLookup.objects.filter(
        backend_path="wbfdm.contrib.dsws.backends.market_data.DSWSMarketDataBackend"
    ):
        try:
            instrument = Instrument.objects.get(id=rel.pms_instrument_id)
            instrument.dl_parameters = {
                "market_data": {
                    "path": "wbfdm.contrib.dsws.dataloaders.market_data.DSWSMarketDataDataloader",
                    "parameters": {
                        "identifier": rel.identifier,
                        "price_symbol": rel.parameters.get("market_data", "P"),
                    },
                },
            }
            instrument.source = "dsws"
            instrument.source_id = rel.identifier
            instrument.save()
        except Exception:
            pass


class Migration(migrations.Migration):
    dependencies = [
        ("geography", "0001_initial"),
        ("wbfdm", "0011_delete_instrumentlookup_instrument_corporate_actions_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="instrument",
            name="dl_parameters",
            field=models.JSONField(blank=True, default=dict),
        ),
        migrations.RemoveField(
            model_name="instrument",
            name="adjustments",
        ),
        migrations.RemoveField(
            model_name="instrument",
            name="corporate_actions",
        ),
        migrations.RemoveField(
            model_name="instrument",
            name="dividends",
        ),
        migrations.RemoveField(
            model_name="instrument",
            name="esg",
        ),
        migrations.RemoveField(
            model_name="instrument",
            name="esg_controversies",
        ),
        migrations.RemoveField(
            model_name="instrument",
            name="financials",
        ),
        migrations.RemoveField(
            model_name="instrument",
            name="market_data",
        ),
        migrations.RemoveField(
            model_name="instrument",
            name="officers",
        ),
        migrations.RemoveField(
            model_name="instrument",
            name="statements",
        ),
        migrations.CreateModel(
            name="InstrumentType",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("name", models.CharField(max_length=128, verbose_name="Name")),
                ("short_name", models.CharField(max_length=128, verbose_name="Short Name")),
                ("key", models.CharField(max_length=32, verbose_name="Key", unique=True)),
                ("description", models.TextField(verbose_name="Description")),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.AlterField(
            model_name="exchange",
            name="city",
            field=models.ForeignKey(
                blank=True,
                help_text="The city where this Exchange is located at.",
                limit_choices_to={"level": 3},
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="exchanges",
                to="geography.geography",
                verbose_name="City",
            ),
        ),
        migrations.RenameField(
            model_name="instrument",
            old_name="instrument_type",
            new_name="old_instrument_type",
        ),
        migrations.AddField(
            model_name="instrument",
            name="instrument_type",
            field=models.ForeignKey(
                default=None,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="instruments",
                to="wbfdm.instrumenttype",
                blank=True,
                null=True,
            ),
            preserve_default=False,
        ),
        migrations.RemoveField(
            model_name="instrumentrequest",
            name="instrument_type",
        ),
        migrations.RunPython(migrate_instrument_type),
        migrations.RemoveField(model_name="instrument", name="old_instrument_type"),
        migrations.AlterField(
            model_name="instrument",
            name="instrument_type",
            field=models.ForeignKey(
                default=None,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="instruments",
                to="wbfdm.instrumenttype",
            ),
            preserve_default=False,
        ),
        migrations.CreateModel(
            name="Equity",
            fields=[],
            options={
                "proxy": True,
                "indexes": [],
                "constraints": [],
            },
            bases=("wbfdm.instrument",),
        ),
        migrations.AddField(
            model_name="instrumentprice",
            name="created",
            field=models.DateTimeField(auto_now_add=True, default=timezone.now(), verbose_name="Created"),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="instrumentprice",
            name="modified",
            field=models.DateTimeField(auto_now=True, verbose_name="Modified"),
        ),
        migrations.RunPython(post_migrate_script),
        migrations.RemoveField(
            model_name="instrumentlookup",
            name="instrument",
        ),
        migrations.DeleteModel(
            name="InstrumentLookup",
        ),
        migrations.RemoveField(
            model_name="instrument",
            name="level",
        ),
        migrations.RemoveField(
            model_name="instrument",
            name="lft",
        ),
        migrations.RemoveField(
            model_name="instrument",
            name="rght",
        ),
        migrations.RemoveField(
            model_name="instrument",
            name="tree_id",
        ),
        migrations.AddField(
            model_name="exchange",
            name="last_updated",
            field=models.DateTimeField(auto_now=True),
        ),
        migrations.AddField(
            model_name="instrument",
            name="employees",
            field=models.IntegerField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name="instrument",
            name="is_security",
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name="instrument",
            name="last_annual_report",
            field=models.DateField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name="instrument",
            name="last_interim_report",
            field=models.DateField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name="instrument",
            name="next_annual_report",
            field=models.DateField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name="instrument",
            name="next_interim_report",
            field=models.DateField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name="instrumenttype",
            name="is_classifiable",
            field=models.BooleanField(default=True, verbose_name="Classifiable"),
        ),
        migrations.AddField(
            model_name="instrumenttype",
            name="is_security",
            field=models.BooleanField(default=True, verbose_name="Security"),
        ),
        migrations.AlterField(
            model_name="controversy",
            name="instrument",
            field=models.ForeignKey(
                limit_choices_to=models.Q(("instrument_type__is_security", True)),
                on_delete=django.db.models.deletion.CASCADE,
                to="wbfdm.instrument",
            ),
        ),
        migrations.AlterField(
            model_name="deal",
            name="equity",
            field=models.ForeignKey(
                limit_choices_to=models.Q(("instrument_type__key", "equity")),
                on_delete=django.db.models.deletion.CASCADE,
                related_name="deals",
                to="wbfdm.instrument",
            ),
        ),
        migrations.AlterField(
            model_name="instrument",
            name="alternative_names",
            field=django.contrib.postgres.fields.ArrayField(
                base_field=models.CharField(blank=True, max_length=255, null=True), blank=True, default=list, size=None
            ),
        ),
        migrations.AlterField(
            model_name="instrument",
            name="description",
            field=models.TextField(blank=True, default="", null=True),
        ),
        migrations.AlterField(
            model_name="instrument",
            name="exchange",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="instruments",
                to="wbfdm.exchange",
            ),
        ),
        migrations.AlterField(
            model_name="instrument",
            name="instrument_type",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="instruments",
                to="wbfdm.instrumenttype",
            ),
        ),
        migrations.AlterField(
            model_name="instrument",
            name="isin",
            field=models.CharField(
                blank=True, help_text="The ISIN provided by the bank.", max_length=12, null=True, verbose_name="ISIN"
            ),
        ),
        migrations.AlterField(
            model_name="instrument",
            name="parent",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="children",
                to="wbfdm.instrument",
                verbose_name="Parent Instrument",
            ),
        ),
        migrations.AlterField(
            model_name="instrument",
            name="refinitiv_identifier_code",
            field=models.CharField(
                blank=True, help_text="Refinitiv Identifier Code", max_length=255, null=True, verbose_name="RIC"
            ),
        ),
        migrations.AlterField(
            model_name="instrument",
            name="refinitiv_mnemonic_code",
            field=models.CharField(
                blank=True,
                help_text="Refinitiv Datastream Mnemonic Code",
                max_length=255,
                null=True,
                verbose_name="Refinitiv Datastream Mnemonic Code",
            ),
        ),
        migrations.AlterField(
            model_name="instrument",
            name="refinitiv_ticker",
            field=models.CharField(
                blank=True, help_text="Refinitiv Refinitiv", max_length=255, null=True, verbose_name="Refinitiv Ticker"
            ),
        ),
        migrations.AlterField(
            model_name="instrumentclassificationrelatedinstrument",
            name="classified_instrument",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="classification_instrument_relationships",
                to="wbfdm.instrumentclassificationthroughmodel",
            ),
        ),
        migrations.AlterField(
            model_name="instrumentclassificationrelatedinstrument",
            name="related_instrument",
            field=models.ForeignKey(
                limit_choices_to=models.Q(("instrument_type__is_classifiable", True)),
                on_delete=django.db.models.deletion.CASCADE,
                related_name="instrument_classification_related",
                to="wbfdm.instrument",
            ),
        ),
        migrations.AlterField(
            model_name="instrumentclassificationthroughmodel",
            name="instrument",
            field=models.ForeignKey(
                limit_choices_to=models.Q(("instrument_type__is_classifiable", True)),
                on_delete=django.db.models.deletion.CASCADE,
                related_name="classifications_through",
                to="wbfdm.instrument",
            ),
        ),
        migrations.AlterField(
            model_name="instrumentclassificationthroughmodel",
            name="related_instruments",
            field=models.ManyToManyField(
                blank=True,
                limit_choices_to=models.Q(("instrument_type__is_classifiable", True)),
                through="wbfdm.InstrumentClassificationRelatedInstrument",
                to="wbfdm.instrument",
            ),
        ),
        migrations.AlterField(
            model_name="instrumentfavoritegroup",
            name="instruments",
            field=models.ManyToManyField(
                blank=True,
                limit_choices_to=models.Q(("children__isnull", True)),
                related_name="favorite_groups",
                to="wbfdm.instrument",
                verbose_name="Favorite Instruments Group",
            ),
        ),
        migrations.AlterField(
            model_name="instrumentlistthroughmodel",
            name="instrument",
            field=models.ForeignKey(
                blank=True,
                limit_choices_to=models.Q(("instrument_type__is_security", True)),
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="wbfdm.instrument",
            ),
        ),
        migrations.AlterField(
            model_name="instrumentprice",
            name="instrument",
            field=models.ForeignKey(
                blank=True,
                limit_choices_to=models.Q(("children__isnull", True)),
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="prices",
                to="wbfdm.instrument",
                verbose_name="Instrument",
            ),
        ),
        migrations.AlterField(
            model_name="instrumenttype",
            name="description",
            field=models.TextField(blank=True, verbose_name="Description"),
        ),
        migrations.AlterField(
            model_name="relatedinstrumentthroughmodel",
            name="instrument",
            field=models.ForeignKey(
                limit_choices_to=models.Q(("children__isnull", True)),
                on_delete=django.db.models.deletion.CASCADE,
                related_name="related_instruments_through",
                to="wbfdm.instrument",
            ),
        ),
        migrations.AlterField(
            model_name="relatedinstrumentthroughmodel",
            name="related_instrument",
            field=models.ForeignKey(
                limit_choices_to=models.Q(("children__isnull", True)),
                on_delete=django.db.models.deletion.CASCADE,
                related_name="dependent_instruments_through",
                to="wbfdm.instrument",
            ),
        ),
        migrations.AddConstraint(
            model_name="exchange",
            constraint=models.UniqueConstraint(fields=("source", "source_id"), name="unique_exchange_source"),
        ),
        migrations.AddField(
            model_name="exchange",
            name="bbg_exchange_codes",
            field=django.contrib.postgres.fields.ArrayField(
                base_field=models.CharField(max_length=4),
                blank=True,
                default=list,
                help_text="Bloomberg Exchange Code.",
                size=None,
                verbose_name="BBG Exchange Code",
            ),
        ),
        migrations.RunPython(migrate_bbg_code),
        migrations.RemoveField(model_name="exchange", name="bbg_exchange"),
        migrations.AddField(
            model_name="instrument",
            name="is_managed",
            field=models.BooleanField(default=False),
        ),
        migrations.AlterField(
            model_name="controversy",
            name="instrument",
            field=models.ForeignKey(
                limit_choices_to=models.Q(("is_security", True)),
                on_delete=django.db.models.deletion.CASCADE,
                to="wbfdm.instrument",
            ),
        ),
        migrations.AlterField(
            model_name="instrumentlistthroughmodel",
            name="instrument",
            field=models.ForeignKey(
                blank=True,
                limit_choices_to=models.Q(("is_security", True)),
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="wbfdm.instrument",
            ),
        ),
    ]



import re
from functools import lru_cache

from wbcore.contrib.currency.models import Currency
from wbcore.contrib.geography.models import Geography

from wbfdm.preferences import get_non_ticker_words

START_DELIMITER = r"(?:^|(?<=[(\s\.\-<]))"
END_DELIMITER = r"(?=[\s\.)\-\>'’,]|$)"


@lru_cache()
def non_ticker_words():
    return (
        list(Geography.countries.values_list("code_2", flat=True))
        + list(Geography.countries.values_list("code_3", flat=True))
        + list(Currency.objects.values_list("key", flat=True))
        + get_non_ticker_words()
    )


def re_ric(input: str):
    # [(\s<]([A-Z]{2}[A-Za-z0-9_.-]+\.[A-Za-z]+)[(\s>] led to too many false positive (e.g. end of sentence when whitespace are missing. We refined by considering only uppercase code
    return set(re.findall(START_DELIMITER + r"([A-Z0-9]{2}[A-Z0-9_.-]+\.[A-Z]+)" + END_DELIMITER, input))


def re_bloomberg(input: str):
    return set(
        filter(
            lambda x: x not in non_ticker_words(),
            re.findall(START_DELIMITER + r"([A-Z]{2,5}(?:\-[A-Z]{2})?)" + END_DELIMITER, input),
        )
    )


def re_isin(input: str):
    return set(re.findall(START_DELIMITER + r"([A-Z]{2}[A-Z0-9]{9}[0-9])" + END_DELIMITER, input))


def re_mnemonic(input: str):
    return set(re.findall(START_DELIMITER + r"([A-Z]+:[A-Z]+)" + END_DELIMITER, input))



from collections import defaultdict
from contextlib import suppress
from datetime import date

import pandas as pd
from wbcore.pandas.utils import (
    override_number_to_percent,
    override_number_to_x,
    override_number_with_currency,
)

from wbfdm.enums import CalendarType, Financial, MarketData, PeriodType
from wbfdm.models import Instrument
from wbfdm.utils import rename_period_index_level_to_repr


class Loader:
    """
    Utility Class to load financial data into a Pandas dataframe with year and interim as index.
    """

    def __init__(
        self,
        instrument: Instrument,
        values: list[Financial],
        calendar_type: CalendarType = CalendarType.FISCAL,
        market_data_values: list[MarketData] | None = None,
        statement_values: list[Financial] | None = None,
        period_type: PeriodType = PeriodType.ALL,
    ):
        self.instrument = instrument
        self.calendar_type = calendar_type
        self.values = values
        self.market_data_values = (
            market_data_values  # specify if any extra market data needs to be merged into the dataframe
        )
        self.statement_values = (
            statement_values  # specify if any extra statement needs to be merged into the dataframe
        )
        self.period_type = period_type
        self.errors: dict[str, list[str]] = defaultdict(list)

    def load(self) -> pd.DataFrame:
        """
        entry function of the class, loads the data into the dataframe and normalize it
        """
        return self._normalize_df(*self._get_base_df())

    def _get_base_df(self) -> tuple[pd.DataFrame, pd.Series]:
        """
        Private method to load data into a dataframe and returns the resulting data with index [year, interim] and "financials" as columns. The source pandas series is returned as second argument

        Returns:
            a tuple of dataframe and series
        """
        # Get the base dataframe from a dataloader
        df = pd.DataFrame(
            Instrument.objects.filter(id=self.instrument.id).dl.financials(
                values=self.values,
                from_year=date.today().year - 5,
                calendar_type=self.calendar_type,
                period_type=self.period_type,
            )
        )
        if df.empty:
            self.errors["missing_data"].append("Missing financial data")
            df = pd.DataFrame(
                columns=[
                    "year",
                    "interim",
                    "period_type",
                    "estimate",
                    "period_end_date",
                    "source",
                    "financial",
                    "value",
                ]
            )

        df["period_end_date"] = pd.to_datetime(df["period_end_date"])
        source_df = (
            df[["year", "interim", "period_type", "source"]].groupby(["year", "interim", "period_type"]).first().source
        )
        # Pivot the data around
        df = df.pivot_table(
            index=["year", "interim", "period_type", "estimate", "period_end_date"],
            columns="financial",
            values="value",
        )
        allowed_columns: list[str] = [v.value for v in self.values]
        if self.statement_values and not df.empty:
            df = self._annotate_statement_data(df, self.statement_values)
            allowed_columns.extend([v.value for v in self.statement_values])
        # If market data should be included here, we need to annotate it and afterwards rename the column axis
        if self.market_data_values and not df.empty:
            df = self._annotate_market_data(df, self.market_data_values)
            allowed_columns.extend([v.value for v in self.market_data_values])
        df = df[[value for value in allowed_columns if value in df.columns]]
        return df, source_df

    # UTILS METHODS

    def _annotate_market_data(self, df: pd.DataFrame, market_data_values: list[MarketData]) -> pd.DataFrame:
        """
        Annotate the given market data into the given dataframe

        Args:
            df: a Pandas dataframe to annotate extra data into
            market_data_values:  a list of MarketData objects

        Returns:
            The extended dataframe with market data
        """
        # We need to reset the indexes as we merge based on columns
        market_data_df = pd.DataFrame(
            Instrument.objects.filter(id=self.instrument.id).dl.market_data(
                from_date=df.index.get_level_values("period_end_date").min() if not df.empty else None,
                target_currency=self.instrument.currency.key,
            )
        )
        values = [mdv.value for mdv in market_data_values]
        # We convert the period_end_date column to a native datetime object to allow for merges based on backwards data
        if not market_data_df.empty:
            market_data_df["period_end_date"] = pd.to_datetime(market_data_df["valuation_date"])
            market_data_df = market_data_df[["period_end_date", *values]].sort_values("period_end_date")
            df = pd.merge_asof(
                left=df.reset_index().sort_values(by="period_end_date"),
                right=market_data_df,
                on="period_end_date",
                direction="backward",
            ).set_index(["year", "interim", "period_type", "estimate", "period_end_date"])
            if df[values].dropna().empty:
                self.errors["missing_data"].append(
                    "We could not find any market data covering the financial statement period"
                )
        return df.rename_axis("financial", axis="columns")

    def _annotate_statement_data(self, df: pd.DataFrame, statement_values: list[Financial]) -> pd.DataFrame:
        """
        Annotate the given statement into the given dataframe

        Args:
            df: a Pandas dataframe to annotate extra data into
            statement_values:  a list of Financial objects

        Returns:
            The extended dataframe with statement
        """
        statement_df = pd.DataFrame(
            Instrument.objects.filter(id=self.instrument.id).dl.statements(
                financials=statement_values, from_year=date.today().year - 5
            ),
        )

        if not statement_df.empty:
            statement_df["period_end_date"] = pd.to_datetime(statement_df["period_end_date"])
            statement_df = statement_df.pivot_table(
                index=["year", "interim", "period_end_date"], columns="financial", values="value"
            )
            statement_df = statement_df.ffill()
            df = pd.merge(
                how="left",
                left=df.reset_index().sort_values(by="period_end_date"),
                right=statement_df,
                on=["year", "interim"],
            ).set_index(["year", "interim", "period_type", "estimate", "period_end_date"])
        else:
            self.errors["missing_data"].append("No statement data")

        return df.rename_axis("financial", axis="columns")

    def _normalize_df(self, df: pd.DataFrame, source_df: pd.Series) -> pd.DataFrame:
        """
        This private method takes a dataframe and it's related source (a source per index), reset the index to be only (year, interim) and detect any possible duplicated row.

        If a duplicate is detected, we appends the attribute `errors` with the duplicated index and take the first one

        Finally, we rename the index into a more human readable format and make sure the the yearly row is present for every year of data

        Args:
            df: the DataFrame to normalize
            source_df: The Series holding the source info for all the DataFrame index

        Returns:
            A normalized DataFrame
        """

        def _ensure_yearly_row_exists(row):
            row["interim"] = row["interim"].astype(int)
            if row[row["interim"] == 0].empty:
                row = pd.concat(
                    [
                        pd.DataFrame(
                            [
                                {
                                    "year": row.name,
                                    "interim": 0,
                                    "period_type": "Y",
                                    "estimate": True,
                                    "period_end_date": row["period_end_date"].max(),
                                }
                            ]
                        ),
                        row,
                    ],
                    axis=0,
                )
            return row

        new_index = (
            df.index.to_frame(index=False)
            .groupby(["year"], group_keys=False, as_index=False)
            .apply(lambda row: _ensure_yearly_row_exists(row), include_groups=True)
            .reset_index(drop=True)
        )
        df = df.reindex(new_index)

        df = df.sort_index().reset_index(
            level=[3, 4], names=["year", "interim", "period_type", "estimate", "period_end_date"]
        )
        # detect duplicates, gracefully handle it by taking the first but log the error for further usage
        index_duplicated = df.index.duplicated()
        if index_duplicated.any():
            for year, interim, period_type in df.index[index_duplicated]:
                interim_info = f"{year} Interim {period_type}{interim}"
                with suppress(KeyError):
                    if source := source_df.loc[(year, interim, period_type), "source"]:
                        interim_info += f" [{source.upper()}]"
                self.errors["duplicated_interims"].append(interim_info)

            # remove duplicated index
            df = df[~index_duplicated]

        df = rename_period_index_level_to_repr(df)

        return df


class FinancialAnalysisResult:
    """
    Wrapper class to help present a multi index pivoted dataframe with (year, interim) as index and financials as columns into a transposed DataFrame for Tree view

    This transposed dataframe is available under the attribute `formatted_df`
    """

    FINANCIAL_MAP = {**MarketData.name_mapping(), **Financial.name_mapping()}

    def __init__(
        self,
        df,
        ordering: list[str] | None = None,
        ignore_group_keys: list[Financial] | None = None,
        override_number_with_currency: str | None = None,
        override_number_with_currency_financials: list[str] | None = None,
        override_number_to_x_financials: list[str] | None = None,
        override_number_to_percent_financials: list[str] | None = None,
        errors: dict[str, list[str]] | None = None,
    ):
        self.df = df
        self.columns = list(self.df.drop(columns=["estimate", "period_end_date"]).columns)

        if ordering:
            ordering.extend(["estimate", "period_end_date"])
            allowed_columns = list(self.FINANCIAL_MAP.keys())
            self.df = self.df[[col for col in ordering if col in self.df.columns]]
            allowed_columns.extend(ordering)
            self.df = self.df.drop(columns=self.df.columns.difference(allowed_columns))

        if not ignore_group_keys:
            ignore_group_keys = []
        self.ignore_group_keys = ignore_group_keys
        self.override_number_with_currency = override_number_with_currency
        self.override_number_with_currency_financials = override_number_with_currency_financials
        self.override_number_to_x_financials = override_number_to_x_financials
        self.override_number_to_percent_financials = override_number_to_percent_financials
        self.errors = errors

        self.formatted_df, self.estimated_mapping = self._get_formatted_df()

    def _get_formatted_df(self) -> tuple[pd.DataFrame, dict[str, bool]]:
        # Transpose and reset the index twice to create an artificial index-column
        df = self.df.copy()

        # Flatten and Rename index into year-interim format
        df.index = df.index.map(lambda index: f"{index[0]}-{index[1]}")

        # store the estimate per index
        estimated_mapping = df["estimate"].to_dict()

        df = df.drop(columns=["estimate", "period_end_date"])

        # get the group keys minus the one ignored
        group_keys = [col if col not in self.ignore_group_keys and col in Financial else None for col in df.columns]

        # Transpose table
        df = df.T.reset_index().reset_index()

        df["_group_key"] = group_keys

        if "financial" in df.columns:
            # set the _overriding columns to define extra decorator for the frontend
            if self.override_number_with_currency and self.override_number_with_currency_financials:
                override_number_with_currency(
                    df,
                    self.override_number_with_currency,
                    *list(map(lambda x: df["financial"] == x, self.override_number_with_currency_financials)),
                )
            if self.override_number_to_x_financials:
                override_number_to_x(
                    df, *list(map(lambda x: df["financial"] == x, self.override_number_to_x_financials))
                )
            if self.override_number_to_percent_financials:
                override_number_to_percent(
                    df, *list(map(lambda x: df["financial"] == x, self.override_number_to_percent_financials))
                )
            # Rename Financials into their verbose representation
            df.financial = df.financial.map(self.FINANCIAL_MAP)

        return df.rename(columns={"index": "id"}), estimated_mapping
